import json
from json import JSONEncoder
import numpy as np
from pymicro.xray.detectors import Detector2d, RegArrayDetector2d
from pymicro.crystal.lattice import Lattice, Symmetry
from pymicro.crystal.microstructure import Microstructure, Grain, Orientation


class ForwardSimulation:
    """Class to represent a Forward Simulation."""

    def __init__(self, sim_type, verbose=False):
        self.sim_type = sim_type
        self.verbose = verbose
        self.exp = Experiment()

    def set_experiment(self, experiment):
        """Attach an X-ray experiment to this simulation."""
        self.exp = experiment


class XraySource:
    """Class to represent a X-ray source."""

    def __init__(self, position=None):
        self.set_position(position)
        self._min_energy = None
        self._max_energy = None

    def set_position(self, position):
        if position is None:
            position = (0., 0., 0.)
        self.position = np.array(position)

    @property
    def min_energy(self):
        return self._min_energy

    @property
    def max_energy(self):
        return self._max_energy

    def set_min_energy(self, min_energy):
        self._min_energy = min_energy

    def set_max_energy(self, max_energy):
        self._max_energy = max_energy

    def set_energy(self, energy):
        """Set the energy (monochromatic case)."""
        self.set_min_energy(energy)
        self.set_max_energy(energy)

    def set_energy_range(self, min_energy, max_energy):
        if min_energy < 0:
            print('specified min energy must be positive, using 0 keV')
            min_energy = 0.
        if max_energy <= min_energy:
            print('specified max energy must be larger than min energy, using %.1f' % min_energy)
        self.set_min_energy(min_energy)
        self.set_max_energy(max_energy)

    def discretize(self, diameter, n=5):
        """Discretize the focus zone of the source into a regular sphere.

        :param float diameter: the diameter of the sphere to use.
        :param int n: the number of point to use alongside the source diameter.
        :return: a numpy array of size (n_inside, 3) with the xyz coordinates of the points inside teh sphere.
        """
        radius = 0.5 * diameter
        step = diameter / n

        # create x, y, z 1D coordinate vectors
        x_source = np.arange(-radius, radius + step, step)
        y_source = np.arange(-radius, radius + step, step)
        z_source = np.arange(-radius, radius + step, step)

        # combine the coordinates in 3D
        xx, yy, zz = np.meshgrid(x_source, y_source, z_source, indexing='ij')

        # filter the coordinate for points inside the sphere
        is_in_sphere = (np.sqrt(xx ** 2 + yy ** 2 + zz ** 2) < radius).astype(np.uint8).ravel()
        inside = np.where(is_in_sphere)[0]

        # assemble the coordinates into a (n^3, 3) array and keep only points inside the sphere
        all_source_positions = np.empty((len(x_source), len(y_source), len(z_source), 3), dtype=float)
        all_source_positions[:, :, :, 0] = xx + self.position[0]
        all_source_positions[:, :, :, 1] = yy + self.position[1]
        all_source_positions[:, :, :, 2] = zz + self.position[2]
        xyz_source = all_source_positions.reshape(-1, all_source_positions.shape[
            -1])  # numpy array with the point coordinates
        xyz_source = xyz_source[inside, :]
        return xyz_source


class SlitsGeometry:
    """Class to represent the geometry of a 4 blades slits."""

    def __init__(self, position=None):
        self.set_position(position)  # central position of the aperture (On Xu direction)
        self.hgap = None  # horizontal opening
        self.vgap = None  # vertical opening

    def set_position(self, position):
        if position is None:
            position = (0., 0., 0.)
        self.position = np.array(position)


class ObjectGeometry:
    """Class to represent any object geometry.
    
    The geometry may have multiple form, including just a point, a regular 3D array or it may be described by a CAD 
    file using the STL format. The array represents the material microstructure using the grain ids and should be set 
    in accordance with an associated  Microstructure instance. In this case, zero represent a non crystalline region.
    """

    def __init__(self, geo_type='point', origin=None):
        self.set_type(geo_type)
        self.set_origin(origin)
        # the positions are initially set at the origin, this is a lazy behaviour as discretizing the volume can be expensive
        self.positions = np.array(self.origin)
        self.array = np.ones((1, 1, 1), dtype=np.uint8)  # unique label set to 1
        self.size = np.array([0., 0., 0.])  # mm units
        self.cad = None

    def set_type(self, geo_type):
        assert (geo_type in ['point', 'array', 'cad']) is True
        self.geo_type = geo_type

    def set_array(self, grain_map, voxel_size):
        self.array = grain_map
        self.size = np.array(grain_map.shape) * voxel_size
        print('size set to {}'.format(self.size))

    def set_origin(self, origin):
        if origin is None:
            origin = (0., 0., 0.)
        self.origin = np.array(origin)

    def get_bounding_box(self):
        if self.geo_type in ['point', 'array']:
            return self.origin - self.size / 2, self.origin + self.size / 2
        elif self.geo_type == 'cad':
            bounds = self.cad.GetBounds()
            return (bounds[0], bounds[2], bounds[4]), (bounds[1], bounds[3], bounds[5])

    def get_positions(self):
        """Return an array of the positions within this sample in world coordinates."""
        return self.positions

    def discretize_geometry(self, grain_id=None):
        """Compute the positions of material points inside the sample.

        A array of size (n_vox, 3) is returned where n_vox is the number of positions. The 3 values of each position
        contain the (x, y, z) coordinates in mm unit. If a grain id is specified, only the positions within this grain
        are returned.

        This is useful in forward simulation where we need to access all the locations within the sample. Three cases
        are available:

         * point Laue diffraction: uses sample origin, grains center and orientation
         * cad Laue diffraction: uses origin, cad file geometry and grains[0] orientation (assumes only one grain)
         * grain map Laue diffraction: uses origin, array, size, and grains orientations.

         to set the cad geometry must be the path to the STL file.
        """
        if self.geo_type == 'point':
            self.positions = np.array(self.origin)
        elif self.geo_type == 'array':
            vx, vy, vz = self.array.shape  # number of voxels
            print(vx, vy, vz)
            bb = self.get_bounding_box()
            print('bounding box is {}'.format(bb))
            x_sample = np.linspace(bb[0][0], bb[1][0], vx)  # mm
            y_sample = np.linspace(bb[0][1], bb[1][1], vy)  # mm
            z_sample = np.linspace(bb[0][2], bb[1][2], vz)  # mm
            if grain_id:
                # filter by the given grain id
                ndx_x, ndx_y, ndx_z = np.where(self.array == grain_id)
                print('found %d voxels in grain %d' % (len(ndx_x), grain_id))
                self.positions = np.c_[x_sample[ndx_x], y_sample[ndx_y], z_sample[ndx_z]]
            else:
                xx, yy, zz = np.meshgrid(x_sample, y_sample, z_sample, indexing='ij')
                all_positions = np.empty((vx, vy, vz, 3), dtype=float)
                all_positions[:, :, :, 0] = xx
                all_positions[:, :, :, 1] = yy
                all_positions[:, :, :, 2] = zz
                self.positions = all_positions.reshape(-1, all_positions.shape[-1])
        elif self.geo_type == 'cad':
            if self.cad is None:
                print('you must set the cad attribute (path to the STL file in mm units) for this geometry')
                return
            from pymicro.view.vtk_utils import is_in_array
            is_in, xyz = is_in_array(self.cad, step=0.2, origin=self.origin)
            self.positions = xyz[np.where(is_in.ravel())]


class Sample:
    """Class to describe a material sample.
    
    A sample is made by a given material (that may have multiple phases), has a name and a position in the experimental 
    local frame. A sample also has a geometry (just a point by default), that may be used to discretize the volume 
    in space or display it in 3D.
    
    .. note::

      For the moment, the material is simply a crystal lattice.
    """

    def __init__(self, name=None, position=None, geo=None, material=None, microstructure=None):
        self.name = name
        self.data_dir = '.'
        self.set_position(position)
        self.set_geometry(geo)
        self.set_material(material)
        self.set_microstructure(microstructure)
        self.grain_ids_path = None

    def set_name(self, name):
        """Set the sample name.
        
        :param str name: The sample name.
        """
        self.name = name

    def set_position(self, position):
        """Set the sample reference position.

        :param tuple position: A vector (tuple or array form) describing the sample position.
        """
        if position is None:
            position = (0., 0., 0.)
        self.position = np.array(position)

    def set_geometry(self, geo):
        """Set the geometry of this sample.

        :param ObjectGeometry geo: A vector (tuple or array form) describing the sample position.
        """
        if geo is None:
            geo = ObjectGeometry()
        assert isinstance(geo, ObjectGeometry) is True
        self.geo = geo

    def get_material(self):
        return self.material

    def set_material(self, material):
        if material is None:
            material = Lattice.cubic(1.0)
        self.material = material

    def get_microstructure(self):
        return self.microstructure

    def set_microstructure(self, microstructure):
        if microstructure is None:
            microstructure = Microstructure()
        self.microstructure = microstructure

    def has_grains(self):
        """Method to see if a sample has at least one grain in the microstructure.
        
        :return: True if the sample has at east one grain, False otherwise."""
        return len(self.microstructure.grains) > 0

    def get_grain_ids(self):
        if not hasattr(self, 'grain_ids'):
            # try to load grain map
            #try:
            import h5py
            import os
            full_path = os.path.join(self.data_dir, self.grain_ids_path)
            print('loading grain_ids field from %s' % full_path)
            f = h5py.File(full_path)
            print('successfully loaded grain_ids with shape {}'.format(f['vol'].shape))
            self.grain_ids = f['vol'][()].transpose(2, 1, 0)  # now we have [x, y, z] representation just like matlab
            f.close()
            #except AttributeError:
            #    print('failed to load grain_ids, please set the grain_ids path for this sample')
        return self.grain_ids

class Experiment:
    """Class to represent an actual or a virtual X-ray experiment.
    
    A cartesian coordinate system (X, Y, Z) is associated with the experiment. By default X is the direction of X-rays 
    and the sample is placed at the origin (0, 0, 0).
    """

    def __init__(self):
        self.source = XraySource()
        self.sample = Sample(name='dummy')
        self.slits = SlitsGeometry()
        self.detectors = []
        self.active_detector_id = -1

    def set_sample(self, sample):
        assert isinstance(sample, Sample) is True
        self.sample = sample

    def get_sample(self):
        return self.sample

    def set_source(self, source):
        assert isinstance(source, XraySource) is True
        self.source = source

    def get_source(self):
        return self.source

    def set_slits(self, slits):
        assert isinstance(slits, SlitsGeometry) is True
        self.slits = slits

    def get_slits(self):
        return self.slits

    def add_detector(self, detector, set_as_active=True):
        """Add a detector to this experiment.
        
        If this is the first detector, the active detector id is set accordingly.

        :param Detector2d detector: an instance of the Detector2d class.
        :param bool set_as_active: set this detector as active.
        """
        assert isinstance(detector, Detector2d) is True
        self.detectors.append(detector)
        if set_as_active:
            self.active_detector_id = self.get_number_of_detectors() - 1

    def get_number_of_detectors(self):
        """Return the number of detector for this experiment."""
        return len(self.detectors)

    def get_active_detector(self):
        """Return the active detector for this experiment."""
        return self.detectors[self.active_detector_id]
    
    def forward_simulation(self, fs, set_result_to_detector=True):
        """Perform a forward simulation of the X-ray experiment onto the active detector.
        
        This typically sets the detector.data field with the computed image.

        :param bool set_result_to_detector: if True, the result is assigned to the current detector.
        :param fs: An instance of `ForwardSimulation` or its derived class.
        """
        if fs.sim_type == 'laue':
            fs.set_experiment(self)
            result = fs.fsim()
        elif fs.sim_type == 'dct':
            fs.set_experiment(self)
            result = fs.dct_projection()
        else:
            print('wrong type of simulation: %s' % fs.sim_type)
            return None
        if set_result_to_detector:
            self.get_active_detector().data = result
        return result

    def save(self, file_path='experiment.txt'):
        """Export the parameters to describe the current experiment to a file using json."""
        dict_exp = {}
        dict_exp['Source'] = self.source
        dict_exp['Sample'] = self.sample
        dict_exp['Detectors'] = self.detectors
        dict_exp['Active Detector Id'] = self.active_detector_id
        # save to file using json
        json_txt = json.dumps(dict_exp, indent=4, cls=ExperimentEncoder)
        with open(file_path, 'w') as f:
            f.write(json_txt)

    @staticmethod
    def load(file_path='experiment.txt'):
        with open(file_path, 'r') as f:
            dict_exp = json.load(f)
        sample = Sample()
        sample.set_name(dict_exp['Sample']['Name'])
        sample.data_dir = dict_exp['Sample']['Data Dir']
        sample.set_position(dict_exp['Sample']['Position'])
        if 'Geometry' in dict_exp['Sample']:
            sample_geo = ObjectGeometry()
            sample_geo.set_type(dict_exp['Sample']['Geometry']['Type'])
            sample.set_geometry(sample_geo)
        if 'Material' in dict_exp['Sample']:
            a, b, c = dict_exp['Sample']['Material']['Lengths']
            alpha, beta, gamma = dict_exp['Sample']['Material']['Angles']
            centering = dict_exp['Sample']['Material']['Centering']
            symmetry = Symmetry.from_string(dict_exp['Sample']['Material']['Symmetry'])
            material = Lattice.from_parameters(a, b, c, alpha, beta, gamma, centering=centering, symmetry=symmetry)
            sample.set_material(material)
        if 'Microstructure' in dict_exp['Sample']:
            micro = Microstructure(dict_exp['Sample']['Microstructure']['Name'])
            # crystal lattice
            if 'Lattice' in dict_exp['Sample']['Microstructure']:
                a, b, c = dict_exp['Sample']['Microstructure']['Lattice']['Lengths']
                alpha, beta, gamma = dict_exp['Sample']['Microstructure']['Lattice']['Angles']
                centering = dict_exp['Sample']['Microstructure']['Lattice']['Centering']
                symmetry = Symmetry.from_string(dict_exp['Sample']['Microstructure']['Lattice']['Symmetry'])
                lattice = Lattice.from_parameters(a, b, c, alpha, beta, gamma, centering=centering, symmetry=symmetry)
                micro.set_lattice(lattice)
            for i in range(len(dict_exp['Sample']['Microstructure']['Grains'])):
                dict_grain = dict_exp['Sample']['Microstructure']['Grains'][i]
                grain = Grain(int(dict_grain['Id']), Orientation.from_euler(dict_grain['Orientation']['Euler Angles (degrees)']))
                grain.position = np.array(dict_grain['Position'])
                grain.volume = dict_grain['Volume']
                if 'hkl_planes' in dict_grain:
                    grain.hkl_planes = dict_grain['hkl_planes']
                micro.grains.append(grain)
            sample.set_microstructure(micro)
        # lazy behaviour, we load only the grain_ids path, the actual array is loaded in memory if needed
        sample.grain_ids_path = dict_exp['Sample']['Grain Ids Path']
        exp = Experiment()
        exp.set_sample(sample)
        source = XraySource()
        source.set_position(dict_exp['Source']['Position'])
        if 'Min Energy (keV)' in dict_exp['Source']:
            source.set_min_energy(dict_exp['Source']['Min Energy (keV)'])
        if 'Max Energy (keV)' in dict_exp['Source']:
            source.set_max_energy(dict_exp['Source']['Max Energy (keV)'])
        exp.set_source(source)
        for i in range(len(dict_exp['Detectors'])):
            dict_det = dict_exp['Detectors'][i]
            if dict_det['Class'] == 'Detector2d':
                det = Detector2d(size=dict_det['Size (pixels)'])
                det.ref_pos = dict_det['Reference Position (mm)']
            if dict_det['Class'] == 'RegArrayDetector2d':
                det = RegArrayDetector2d(size=dict_det['Size (pixels)'])
                det.pixel_size = dict_det['Pixel Size (mm)']
                det.ref_pos = dict_det['Reference Position (mm)']
                if 'Min Energy (keV)' in dict_exp['Detectors']:
                    det.tilt = dict_det['Tilts (deg)']
                if 'Binning' in dict_det:
                    det.set_binning(dict_det['Binning'])
                det.u_dir = np.array(dict_det['u_dir'])
                det.v_dir = np.array(dict_det['v_dir'])
                det.w_dir = np.array(dict_det['w_dir'])
            exp.add_detector(det)
        return exp


class ExperimentEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, ObjectGeometry):
            dict_geo = {}
            dict_geo['Type'] = o.geo_type
            return dict_geo
        if isinstance(o, Lattice):
            dict_lattice = {}
            dict_lattice['Angles'] = o._angles.tolist()
            dict_lattice['Lengths'] = o._lengths.tolist()
            dict_lattice['Centering'] = o._centering
            dict_lattice['Symmetry'] = o._symmetry.to_string()
            return dict_lattice
        if isinstance(o, Sample):
            dict_sample = {}
            dict_sample['Name'] = o.name
            dict_sample['Data Dir'] = o.data_dir
            dict_sample['Position'] = o.position.tolist()
            dict_sample['Geometry'] = o.geo
            dict_sample['Material'] = o.material
            dict_sample['Microstructure'] = o.microstructure
            dict_sample['Grain Ids Path'] = o.grain_ids_path
            return dict_sample
        if isinstance(o, RegArrayDetector2d):
            dict_det = {}
            dict_det['Class'] = o.__class__.__name__
            dict_det['Size (pixels)'] = o.size
            dict_det['Pixel Size (mm)'] = o.pixel_size
            dict_det['Data Type'] = str(o.data_type)
            dict_det['Reference Position (mm)'] = o.ref_pos.tolist()
            dict_det['Binning'] = o.binning
            dict_det['u_dir'] = o.u_dir.tolist()
            dict_det['v_dir'] = o.v_dir.tolist()
            dict_det['w_dir'] = o.w_dir.tolist()
            return dict_det
        if isinstance(o, Detector2d):
            dict_det = {}
            dict_det['Class'] = o.__class__.__name__
            dict_det['Size (pixels)'] = o.size
            dict_det['Data Type'] = o.data_type
            dict_det['Reference Position (mm)'] = o.ref_pos.tolist()
            return dict_det
        if isinstance(o, XraySource):
            dict_source = {}
            dict_source['Position'] = o.position.tolist()
            if o.min_energy is not None:
                dict_source['Min Energy (keV)'] = o.min_energy
            if o.max_energy is not None:
                dict_source['Max Energy (keV)'] = o.max_energy
            return dict_source
        if isinstance(o, Microstructure):
            dict_micro = {}
            dict_micro['Name'] = o.name
            dict_micro['Lattice'] = o.get_lattice()
            dict_micro['Grains'] = o.grains
            return dict_micro
        if isinstance(o, Grain):
            dict_grain = {}
            dict_grain['Id'] = float(o.id)
            dict_grain['Position'] = o.center.tolist()
            dict_grain['Orientation'] = o.orientation
            dict_grain['Volume'] = o.volume
            if hasattr(o, 'hkl_planes'):
                dict_grain['hkl_planes'] = o.hkl_planes
            return dict_grain
        if isinstance(o, Orientation):
            dict_orientation = {}
            dict_orientation['Euler Angles (degrees)'] = o.euler.tolist()
            return dict_orientation






'''The fitting module define some classes to easily perform 1D curve 
   fitting. This module supports fitting (x, y) data with a general 
   mechanism. Any fitting function can be provided to the fit method 
   and a few general purpose fuctions are predefined:
   
    * Gaussian
    * Lorentzian
    * Cosine
    * Voigt

  .. figure:: _static/fitting_functions.png
      :width: 400 px
      :height: 300 px
      :alt: lattice_3d
      :align: center

      Plot of the main predefined fitting function in the fitting module.
'''
from scipy import optimize
from scipy.special import wofz
import numpy as np


def fit(y, x=None, expression=None, nb_params=None, init=None):
    '''Static method to perform curve fitting directly.

      *Parameters*

      **y**: the data to match (a 1d numpy array)

      **x**: the corresponding x coordinates (optional, None by default)

      **expression**: can be either a string to select a predefined
      function or alternatively a user defined function with the signature
      f(x, p) (in this case you must specify the length of the parameters
      array p via setting nb_params).

      **nb_params**: the number of parameters of the user defined fitting
      function (only needed when a custom fitting function is provided,
      None by default)

      **init**: a sequence (the length must be equal to the number of
      parameters of the fitting function) used to initialise the fitting
      function.

      For instance, to fit some (x,y) data with a gaussian function, simply use:
       ::

         F = fit(y, x, expression='Gaussian')

       Alternatively you may specify you own function directly defined with Python, like:
       ::

         def myf(x, p):
           return p[0]*x + p[1]

         F = fit(y, x, expression=myf, nb_params=2)
    '''
    if expression == 'Gaussian':
        F = Gaussian()
    elif expression == 'Lorentzian':
        F = Lorentzian()
    elif expression == 'Cosine':
        F = Cosine()
    elif expression == 'Voigt':
        F = Voigt()
    else:
        F = FitFunction()
        if not nb_params:
            print('please specify the number of parameters for your fit function, aborting fit...')
            return None
        if not init:
            init = np.ones(nb_params)
        if not len(init) == nb_params:
            print(
                'there are more parameters in the fit function than specified in the initialization sequence, aborting initialization...')
            init = np.ones(nb_params)
        for i in range(nb_params):
            F.add_parameter(init[i], 'p%d' % i)
        F.expression = expression
    F.fit(y, x)
    return F


def lin_reg(xi, yi):
    """Apply linear regression to a series of points.
    
    This function return the best linear fit in the least square sense.
    :param ndarray xi: a 1D array of the x coordinate.
    :param ndarray yi: a 1D array of the y coordinate.
    :return tuple: the linear intercept, slope and correlation coefficient.
    """
    n = len(xi)
    assert (n == len(yi))
    sx = np.sum(xi)
    sy = np.sum(yi)
    sxx = np.sum(xi ** 2)
    sxy = np.sum(xi * yi)
    syy = np.sum(yi ** 2)
    beta = (n * sxy - sx * sy) / (n * sxx - sx ** 2)
    alpha = 1. / n * sy - 1. / n * sx * beta
    r = (n * sxy - sx * sy) / np.sqrt((n * sxx - sx ** 2) * (n * syy - sy ** 2))
    return alpha, beta, r


class Parameter:
    '''A class to handle modiable parameters.'''

    def __init__(self, value, name=None):
        '''Create a new parameter with the given value.'''
        self.value = value
        self.name = name

    def set(self, value):
        '''Set the value of the parameter.'''
        self.value = value

    def set_name(self, name):
        '''Set the name of the parameter.'''
        self.name = name

    def __call__(self):
        '''With this we can use p() if p is an instance of Parameter.'''
        return self.value

    def __repr__(self):
        '''Provide a string representation of the parameter, simply based
        on its actual value.'''
        return str(self.value)


class FitFunction:
    '''This class provides a basic canvas to define a fit function.

    You may subclass it to create your own fitting function just as the
    predfined fit function do (see `Gaussian` for instance).
    '''

    def __init__(self):
        self.parameters = []
        self.expression = None

    def get_parameters(self):
        '''Return the list of parameters of this fit function.'''
        return self.parameters

    def get_parameter_names(self):
        '''Return the list os parameter names of this fit function.'''
        names = []
        for p in self.get_parameters:
            names.append(p.name)
        return names

    def add_parameter(self, value, name):
        param = Parameter(value, name)
        self.parameters.append(param)

    def __call__(self, x):
        '''With this we can call directly f(x) to evaluate the function f at x.'''
        return self.compute(x)

    def __repr__(self):
        '''Provide a string representation of the fitting function, giving
        its type and the list of its parameters with names and values.'''
        s = '%s function\n' % self.__class__.__name__
        s += 'Parameters are:\n'
        params = self.get_parameters()
        for i in range(len(params)):
            s += ' * %s = %g\n' % (params[i].name, params[i].value)
        return s

    def compute(self, x):
        '''Evaluate the fit function at coordinates x.'''
        p = self.get_parameters()
        return self.expression(x, p)

    def fit(self, y, x=None, verbose=False):
        '''Perform fitting on the given data.

        This will adjust the parameters of this fit function to match as
        well as possible the given data using a least square minimisation.

        *Parameters*

        **y**: the data to match (a 1d numpy array)

        **x**: the corresponding x coordinates (optional, None by default)

        **verbose**: boolean, activate verbose mode
        '''
        # local function to evaluate the cost
        global it
        it = 0

        def cost_func(new_params):
            global it
            p = self.get_parameters()
            if verbose:
                print('iteration %d, trying parameters:' % it, p)
                it += 1
                '''
                from matplotlib import pyplot as plt
                if it == 0:
                  plt.plot(x, y, 'bo', label = 'data points')
                  plt.ylim(0, 0.6)
                  plt.grid()
                  plt.legend(numpoints=1,loc='upper left')
                  plt.savefig('fit/fit_%02d.pdf' % it)
                it += 1
                plt.clf()
                plt.plot(x, y, 'bo', label = 'data points')
                plt.plot(x, self(x), 'k-', label = 'gaussian fit')
                plt.ylim(0, 0.6)
                plt.title('fitting iteration %02d' % it)
                plt.legend(numpoints=1,loc='upper left')
                plt.savefig('fit/fit_%02d.pdf' % it)
                '''
            for i, pi in enumerate(p):
                pi.set(new_params[i])
            return y - self(x)

        if x is None: x = np.arange(y.shape[0])
        p = [param.value for param in self.get_parameters()]
        optimize.leastsq(cost_func, p, Dfun=None, xtol=1.e-6)

class SumOfFitFunction(FitFunction):

    def __init__(self, function_list):
        self.parameters = []
        for f in function_list:
            self.parameters.extend(f.get_parameters())
        self.expression = function_list

    def __repr__(self):
        '''Provide a string representation of the fitting function, giving
        its type and the list of its parameters with names and values.'''
        s = '%s function\n' % self.__class__.__name__
        s += 'list of function in the sum:'
        for f in self.expression:
            s += f.__repr__()
        return s

    def compute(self, x):
        '''Evaluate the fit function at coordinates x.'''
        result = np.zeros_like(x)
        for f in self.expression:
            result += f.compute(x)
        return result


class Gaussian(FitFunction):
    '''first parameter is position, second is sigma, third is height'''

    def __init__(self, position=0.0, sigma=1.0, height=1.0):
        FitFunction.__init__(self)

        def G(x, p):
            return p[2].value * np.exp(-((x - p[0].value) / p[1].value) ** 2)

        self.expression = G
        self.add_parameter(position, 'position')
        self.add_parameter(sigma, 'sigma')
        self.add_parameter(height, 'height')

    def set_position(self, position):
        '''Set the position (center) of the gauss function.'''
        self.parameters[0].set(position)

    def set_sigma(self, sigma):
        '''Set the width (variance) of the gauss function.'''
        self.parameters[1].set(sigma)

    def set_height(self, height):
        '''Set the maximum (height) of the gauss function.'''
        self.parameters[2].set(height)

    def fwhm(self):
        '''Compute the full width at half maximum of the gauss function.'''
        p = self.get_parameters()
        return 2 * p[1].value * np.sqrt(np.log(2))


class Lorentzian(FitFunction):
    '''Lorentzian funtion.

    The first parameter is the position, the second is gamma. The maximum
    of the function is given by height_factor/(pi*gamma). The FWHM is just 2*gamma.
    '''

    def __init__(self, position=0.0, gamma=1.0, height_factor=1.0):
        FitFunction.__init__(self)

        def L(x, p):
            return p[2].value * p[1].value / np.pi / ((x - p[0].value) ** 2 + p[1].value ** 2)

        self.expression = L
        self.add_parameter(position, 'position')
        self.add_parameter(gamma, 'width')
        self.add_parameter(height_factor, 'height_factor')

    def set_position(self, position):
        self.parameters[0].set(position)

    def set_gamma(self, gamma):
        self.parameters[1].set(gamma)

    def set_height(self, height):
        '''Set the maximum (height) of the Lorentzian function. This
        actually set the height factor to the value height*pi*gamma.
        '''
        self.parameters[2].set(height * np.pi * self.parameters[1].value)

    def fwhm(self):
        p = self.get_parameters()
        return 2 * p[1].value


class Cosine(FitFunction):
    '''first parameter is position, second is width'''

    def __init__(self, position=0.0, width=1.0):
        FitFunction.__init__(self)

        def C(x, p):
            return np.cos(np.pi * (x - p[0].value) / (2 * p[1].value))

        self.expression = C
        self.add_parameter(position, 'position')
        self.add_parameter(width, 'width')

    def set_position(self, position):
        self.parameters[0].set(position)

    def set_width(self, a):
        self.parameters[1].set(a)

    def fwhm(self):
        p = self.get_parameters()
        return 3. / 4 * p[1].value


class Voigt(FitFunction):
    '''The Voigt function is also the real part of
    w(x) = exp(-x**2) erfc(ix), the Faddeeva function.

    Here we use one of the popular implementation which is available
    in scipy with the wofz function.
    '''

    def __init__(self, position=0.0, sigma=1.0, gamma=1.0, height_factor=1.0):
        FitFunction.__init__(self)

        def V(x, p):
            z = (x - p[0].value + 1j * p[2].value) / (p[1].value * np.sqrt(2))
            return p[3].value * wofz(z).real / (p[1].value * np.sqrt(2 * np.pi))

        self.expression = V
        self.add_parameter(position, 'position')
        self.add_parameter(sigma, 'sigma')
        self.add_parameter(gamma, 'gamma')
        self.add_parameter(height_factor, 'height_factor')

    def set_position(self, position):
        '''Set the position (center) of the Voigt function.'''
        self.parameters[0].set(position)

    def set_sigma(self, sigma):
        '''Set the sigma of the Voigt function.'''
        self.parameters[1].set(sigma)

    def set_height(self, height):
        '''Set the maximum (height) of the Voigt function. This
        actually set the height factor to the proper value. Be careful that
        if you change the other parameters (sigma, gamma) the maximum height
        will be changed.
        '''
        maxi = self.compute(self.parameters[0].value)
        self.parameters[3].set(height / maxi)

    def fwhm(self):
        '''Compute the full width at half maximum of the Voigt function.

        The height factor does not change the fwhm. The fwhm can be evaluated
        by the width of the associated Gaussian and Lorentzian functions.
        The fwhm is approximated by the equation from J. Olivero and R. Longbothum [1977]'''
        p = self.get_parameters()
        fg = 2 * p[1].value * np.sqrt(2 * np.log(2))
        fl = 2 * p[2].value
        return 0.5346 * fl + np.sqrt(0.2166 * fl ** 2 + fg ** 2)



#!/usr/bin/env python
import sys
import os
import re

import numpy as np
from matplotlib.figure import Figure
import matplotlib.image as mpimg

from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT as NavigationToolbar

from PyQt5.QtWidgets import QApplication, QWidget, QMainWindow, QHBoxLayout, QVBoxLayout, QComboBox, QLabel, QCheckBox
from PyQt5.QtCore import Qt

from pymicro.file.file_utils import edf_read
from pymicro.external.tifffile import TiffFile
from pymicro.view.vol_utils import hist


class PlotWidget(QWidget):
    def __init__(self, parent=None, image=None, toolbar=True):
        QWidget.__init__(self, parent)
        self.data = image
        self.fliplr = False
        self.flipud = False
        self.dpi = 100
        self.cmap = 'gray'
        self.toolbar = toolbar
        self.create_main_widget()
        if self.data is not None:
            self.on_draw()

    def create_main_widget(self):
        print(self.data)
        self.fig = Figure((10.0, 8.0), dpi=self.dpi)
        self.axes = self.fig.add_subplot(111)
        self.canvas = FigureCanvas(self.fig)
        self.canvas.setFocusPolicy(Qt.StrongFocus)
        self.canvas.setFocus()
        vbox = QVBoxLayout()
        vbox.addWidget(self.canvas)  # the matplotlib canvas
        if self.toolbar:
            self.mpl_toolbar = NavigationToolbar(self.canvas, self)
            vbox.addWidget(self.mpl_toolbar)
        self.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.setLayout(vbox)

    def set_image(self, image):
        self.data = image
        self.on_draw()

    def set_fliplr(self, state):
        self.fliplr = state
        self.on_draw()

    def set_flipud(self, state):
        self.flipud = state
        self.on_draw()

    def on_key_press(self, event):
        if event.key == 'i':
            print('* image infos:')
            print('shape is (%dx%d)' % (np.shape(self.data)[0], np.shape(self.data)[1]))
            print('min in image= %d, max in image=%d' % (np.min(self.data), np.max(self.data)))
        elif event.key == 'right':
            # load next image
            self.parent().parent().on_next_image()
        elif event.key == 'left':
            # load previous image
            self.parent().parent().on_prev_image()
        elif event.key == 'h':
            # plot the image histogram
            hist(self.data, data_range=(np.min(self.data), np.max(self.data)), show=True)
        # implement the default mpl key press events described at
        # http://matplotlib.org/users/navigation_toolbar.html#navigation-keyboard-shortcuts
        key_press_handler(event, self.canvas, self.mpl_toolbar)

    def on_draw(self):
        # andle image flips
        data = self.data
        if self.fliplr:
            data = np.fliplr(data)
        if self.flipud:
            data = np.flipud(data)
        self.axes.imshow(data, cmap=self.cmap, origin='upper', interpolation='nearest',
                         clim=[np.min(self.data), np.max(self.data)])
        self.canvas.draw()


class ImageViewerForm(QMainWindow):
    def __init__(self, parent=None, images=None, image_type='path'):
        QMainWindow.__init__(self, parent)
        self.image_type = image_type  # can be either path or data
        self.images = images
        self.index = 0
        self.create_main_frame(self.get_image_data())
        self.plot_widget.on_draw()

    def get_image_number(self):
        if self.image_type == 'path':
            return len(self.images)
        elif self.image_type == 'data':
            return self.images.shape[2]

    def create_main_frame(self, image):
        print('create_main_frame')
        self.main_frame = QWidget()
        vbox = QVBoxLayout()

        # create the plot widget
        self.plot_widget = PlotWidget(image=image, parent=self, toolbar=True)
        vbox.addWidget(self.plot_widget)

        # create the flip check boxes bar
        hbox_flips = QHBoxLayout()
        cblr = QCheckBox('Flip L/R', self)
        cblr.stateChanged.connect(self.plot_widget.set_fliplr)
        hbox_flips.addWidget(cblr)
        cbud = QCheckBox('Flip U/D', self)
        hbox_flips.addWidget(cbud)
        cbud.stateChanged.connect(self.plot_widget.set_flipud)
        vbox.addLayout(hbox_flips)

        # create the color map bar
        hbox = QHBoxLayout()
        self.cmap_label = QLabel('Image color map')
        hbox.addWidget(self.cmap_label)
        self.cmap_combo = QComboBox()
        self.cmap_combo.addItem('gray')
        self.cmap_combo.addItem('jet')
        self.cmap_combo.addItem('hot')
        self.cmap_combo.activated[str].connect(self.on_cmap_selected)
        hbox.addWidget(self.cmap_combo)
        vbox.addLayout(hbox)
        self.main_frame.setLayout(vbox)
        self.setCentralWidget(self.main_frame)

    def on_cmap_selected(self, text):
        self.plot_widget.cmap = str(text)
        self.plot_widget.on_draw()

    def on_prev_image(self):
        self.index -= 1
        if self.index < 0: self.index += self.get_image_number()
        # get the previous image
        data = self.get_image_data()
        self.plot_widget.set_image(data)

    def on_next_image(self):
        self.index = (self.index + 1) % self.get_image_number()
        # get the next image
        data = self.get_image_data()
        self.plot_widget.set_image(data)

    def get_image_data(self):
        if self.image_type == 'path':
            data = self.load_image(self.images[self.index])
        elif self.image_type == 'data':
            data = self.images[:, :, self.index].T
        return data

    def load_image(self, image_path):
        print('loading image from %s' % image_path)
        # read image depending on file extension
        if image_path.endswith('.edf'):
            image = edf_read(image_path).transpose()
        elif image_path.endswith('.png'):
            image = mpimg.imread(image_path)
        elif image_path.endswith('.tif'):
            image = TiffFile(image_path).asarray()  # should we transpose here?
        else:
            print('Only png, tif and edf images are supported for the moment')
            image = np.zeros((5, 5), dtype=np.uint8)  # dummy image
        return image


class ImageViewer(QApplication):
    def __init__(self, args):
        '''Init the ImageViewer application.'''
        app = QApplication(['Pymicro Image Viewer'])
        # print(args)
        print(type(args[0]))
        # check if first arg is a numpy array
        if type(args) == np.ndarray:  # this may be replaced by a 'data' keyword...
            form = ImageViewerForm(images=args, image_type='data')
        else:
            # check if first arg contains '='
            if args[0].find('=') < 0:
                start = 1  # app launched from the command line, first arg is just the name of the file
            else:
                start = 0
            # parse the list of args into a dictionary
            d = dict(map(lambda x: x.split('='), args[start:]))
            print('received args:', d)
            if 'wdir' in d:
                self.wdir = d['wdir']
            else:
                self.wdir = '.'
            if 'pattern' in d:
                pattern = d['pattern']
            else:
                pattern = '.png$'
            image_list = []
            file_list = os.listdir(self.wdir)
            file_list.sort()
            for f in file_list:
                if re.search(pattern, f):
                    image_list.append(os.path.join(self.wdir, f))
            if not image_list:
                print('No image found, please verify your pattern or your working directory')
                sys.exit(1)
            form = ImageViewerForm(images=image_list, image_type='path')
        form.show()
        app.exec_()


if __name__ == "__main__":
    ImageViewer(['wdir=/home/proudhon/data/tomo/rawdata/ma1921/Pb_balls_316LN_', 'pattern=.edf'])
    #ImageViewer(sys.argv)
    # a = np.zeros((100, 50, 10))
    # a[50, :, 1] = 1
    # ImageViewer(a)



"""A package to specifically handle crystal structures and microstructures.

   The crystal package has three main modules:

   * :py:mod:`~pymicro.crystal.lattice`
   * :py:mod:`~pymicro.crystal.microstructure`
   * :py:mod:`~pymicro.crystal.texture`

   .. moduleauthor:: Henry Proudhon <henry.proudhon@gmail.com>

"""



"""The lattice module define the class to handle 3D crystal lattices (the 14 Bravais lattices).
"""
import os
from pymicro.external import CifFile_module as CifFile
import enum
import functools
import numpy as np
from numpy import pi, dot, transpose, radians
from matplotlib import pyplot as plt


class Crystal:
    '''
    The Crystal class to create any particular crystal structure.

    A crystal instance is composed by:

     * one of the 14 Bravais lattice
     * a point basis (or motif)
    '''

    def __init__(self, lattice, basis=None, basis_labels=None, basis_sizes=None, basis_colors=None):
        '''
        Create a Crystal instance with the given lattice and basis.

        This create a new instance of a Crystal object. The given lattice
        is assigned to the crystal. If the basis is not specified, it will
        be one atom at (0., 0., 0.).

        :param lattice: the :py:class:`~pymicro.crystal.lattice.Lattice` instance of the crystal.
        :param list basis: A list of tuples containing the position of the atoms in the motif.
        :param list basis_labels: A list of strings containing the description of the atoms in the motif.
        :param list basis_labels: A list of float between 0. and 1. (default 0.1) to sale the atoms in the motif.
        :param list basis_colors: A list of vtk colors of the atoms in the motif.
        '''
        self._lattice = lattice
        if basis == None:
            # default to one atom at (0, 0, 0)
            self._basis = [(0., 0., 0.)]
            self._labels = ['?']
            self._sizes = [0.1]
            self._colors = [(0., 0., 1.)]
        else:
            self._basis = basis
            self._labels = basis_labels
            self._sizes = basis_sizes
            self._colors = basis_colors


class Symmetry(enum.Enum):
    """
    Class to describe crystal symmetry defined by its Laue class symbol.
    """
    cubic = 'm3m'
    hexagonal = '6/mmm'
    orthorhombic = 'mmm'
    tetragonal = '4/mmm'
    trigonal = 'bar3m'
    monoclinic = '2/m'
    triclinic = 'bar1'

    @staticmethod
    def from_string(s):
        if s == 'cubic':
            return Symmetry.cubic
        elif s == 'hexagonal':
            return Symmetry.hexagonal
        elif s == 'orthorhombic':
            return Symmetry.orthorhombic
        elif s == 'tetragonal':
            return Symmetry.tetragonal
        elif s == 'trigonal':
            return Symmetry.trigonal
        elif s == 'monoclinic':
            return Symmetry.monoclinic
        elif s == 'triclinic':
            return Symmetry.triclinic
        else:
            return None

    def to_string(self):
        if self is Symmetry.cubic:
            return 'cubic'
        elif self is Symmetry.hexagonal:
            return 'hexagonal'
        elif self is Symmetry.orthorhombic:
            return 'orthorhombic'
        elif self is Symmetry.tetragonal:
            return 'tetragonal'
        elif self is Symmetry.trigonal:
            return 'trigonal'
        elif self is Symmetry.monoclinic:
            return 'monoclinic'
        elif self is Symmetry.triclinic:
            return 'triclinic'
        else:
            return None

    def symmetry_operators(self, use_miller_bravais=False):
        """Define the equivalent crystal symmetries.

        Those come from Randle & Engler, 2000. For instance in the cubic
        crystal struture, for instance there are 24 equivalent cube orientations.

        :returns array: A numpy array of shape (n, 3, 3) where n is the \
        number of symmetries of the given crystal structure.
        """
        if self is Symmetry.cubic:
            sym = np.zeros((24, 3, 3), dtype=np.float)
            sym[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
            sym[1] = np.array([[0., 0., -1.], [0., -1., 0.], [-1., 0., 0.]])
            sym[2] = np.array([[0., 0., -1.], [0., 1., 0.], [1., 0., 0.]])
            sym[3] = np.array([[-1., 0., 0.], [0., 1., 0.], [0., 0., -1.]])
            sym[4] = np.array([[0., 0., 1.], [0., 1., 0.], [-1., 0., 0.]])
            sym[5] = np.array([[1., 0., 0.], [0., 0., -1.], [0., 1., 0.]])
            sym[6] = np.array([[1., 0., 0.], [0., -1., 0.], [0., 0., -1.]])
            sym[7] = np.array([[1., 0., 0.], [0., 0., 1.], [0., -1., 0.]])
            sym[8] = np.array([[0., -1., 0.], [1., 0., 0.], [0., 0., 1.]])
            sym[9] = np.array([[-1., 0., 0.], [0., -1., 0.], [0., 0., 1.]])
            sym[10] = np.array([[0., 1., 0.], [-1., 0., 0.], [0., 0., 1.]])
            sym[11] = np.array([[0., 0., 1.], [1., 0., 0.], [0., 1., 0.]])
            sym[12] = np.array([[0., 1., 0.], [0., 0., 1.], [1., 0., 0.]])
            sym[13] = np.array([[0., 0., -1.], [-1., 0., 0.], [0., 1., 0.]])
            sym[14] = np.array([[0., -1., 0.], [0., 0., 1.], [-1., 0., 0.]])
            sym[15] = np.array([[0., 1., 0.], [0., 0., -1.], [-1., 0., 0.]])
            sym[16] = np.array([[0., 0., -1.], [1., 0., 0.], [0., -1., 0.]])
            sym[17] = np.array([[0., 0., 1.], [-1., 0., 0.], [0., -1., 0.]])
            sym[18] = np.array([[0., -1., 0.], [0., 0., -1.], [1., 0., 0.]])
            sym[19] = np.array([[0., 1., 0.], [1., 0., 0.], [0., 0., -1.]])
            sym[20] = np.array([[-1., 0., 0.], [0., 0., 1.], [0., 1., 0.]])
            sym[21] = np.array([[0., 0., 1.], [0., -1., 0.], [1., 0., 0.]])
            sym[22] = np.array([[0., -1., 0.], [-1., 0., 0.], [0., 0., -1.]])
            sym[23] = np.array([[-1., 0., 0.], [0., 0., -1.], [0., -1., 0.]])
        elif self is Symmetry.hexagonal:
            if use_miller_bravais:
              # using the Miller-Bravais representation here
              sym = np.zeros((12, 4, 4), dtype=np.int)
              sym[0] = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
              sym[1] = np.array([[0, 0, 1, 0], [1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
              sym[2] = np.array([[0, 1, 0, 0], [0, 0, 1, 0], [1, 0, 0, 0], [0, 0, 0, 1]])
              sym[3] = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])
              sym[4] = np.array([[0, 0, 1, 0], [1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1]])
              sym[5] = np.array([[0, 1, 0, 0], [0, 0, 1, 0], [1, 0, 0, 0], [0, 0, 0, -1]])
              sym[6] = np.array([[-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
              sym[7] = np.array([[0, 0, -1, 0], [-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
              sym[8] = np.array([[0, -1, 0, 0], [0, 0, -1, 0], [-1, 0, 0, 0], [0, 0, 0, 1]])
              sym[9] = np.array([[-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, -1]])
              sym[10] = np.array([[0, 0, -1, 0], [-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, -1]])
              sym[11] = np.array([[0, -1, 0, 0], [0, 0, -1, 0], [-1, 0, 0, 0], [0, 0, 0, -1]])
            else:
              sym = np.zeros((12, 3, 3), dtype=np.float)
              s60 = np.sin(60 * np.pi / 180)
              sym[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
              sym[1] = np.array([[0.5, s60, 0.], [-s60, 0.5, 0.], [0., 0., 1.]])
              sym[2] = np.array([[-0.5, s60, 0.], [-s60, -0.5, 0.], [0., 0., 1.]])
              sym[3] = np.array([[-1., 0., 0.], [0., -1., 0.], [0., 0., 1.]])
              sym[4] = np.array([[-0.5, -s60, 0.], [s60, -0.5, 0.], [0., 0., 1.]])
              sym[5] = np.array([[0.5, -s60, 0.], [s60, 0.5, 0.], [0., 0., 1.]])
              sym[6] = np.array([[1., 0., 0.], [0., -1., 0.], [0., 0., -1.]])
              sym[7] = np.array([[0.5, s60, 0.], [s60, -0.5, 0.], [0., 0., -1.]])
              sym[8] = np.array([[-0.5, s60, 0.], [s60, 0.5, 0.], [0., 0., -1.]])
              sym[9] = np.array([[-1., 0., 0.], [0., 1., 0.], [0., 0., -1.]])
              sym[10] = np.array([[-0.5, -s60, 0.], [-s60, 0.5, 0.], [0., 0., -1.]])
              sym[11] = np.array([[0.5, -s60, 0.], [-s60, -0.5, 0.], [0., 0., -1.]])
        elif self is Symmetry.orthorhombic:
            sym = np.zeros((4, 3, 3), dtype=np.float)
            sym[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
            sym[1] = np.array([[1., 0., 0.], [0., -1., 0.], [0., 0., -1.]])
            sym[2] = np.array([[-1., 0., -1.], [0., 1., 0.], [0., 0., -1.]])
            sym[3] = np.array([[-1., 0., 0.], [0., -1., 0.], [0., 0., 1.]])
        elif self is Symmetry.tetragonal:
            sym = np.zeros((8, 3, 3), dtype=np.float)
            sym[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
            sym[1] = np.array([[0., -1., 0.], [1., 0., 0.], [0., 0., 1.]])
            sym[2] = np.array([[-1., 0., 0.], [0., -1., 0.], [0., 0., 1.]])
            sym[3] = np.array([[0., 1., 0.], [-1., 0., 0.], [0., 0., 1.]])
            sym[4] = np.array([[1., 0., 0.], [0., -1., 0.], [0., 0., -1.]])
            sym[5] = np.array([[-1., 0., 0.], [0., 1., 0.], [0., 0., -1.]])
            sym[6] = np.array([[0., 1., 0.], [1., 0., 0.], [0., 0., -1.]])
            sym[7] = np.array([[0., -1., 0.], [-1., 0., 0.], [0., 0., -1.]])
        elif self is Symmetry.triclinic:
            sym = np.zeros((1, 3, 3), dtype=np.float)
            sym[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
        else:
            raise ValueError('warning, symmetry not supported: %s' % self)
        return sym

    def move_rotation_to_FZ(self, g, verbose=False):
        """
        Compute the rotation matrix in the Fundamental Zone of a given `Symmetry` instance.

        :param g: a 3x3 matrix representing the rotation 
        :param verbose: flag for verbose mode
        :return: a new 3x3 matrix for the rotation in the fundamental zone.
        """
        omegas = []  # list to store all the rotation angles
        syms = self.symmetry_operators()
        for sym in syms:
            # apply the symmetry operator
            om = np.dot(sym, g)
            if verbose:
                print(om)
                print(om.trace())
            # compute the Rodrigues vector of the corresponding orientation matrix
            # from pymicro.crystal.microstructure import Orientation
            # r = Orientation.OrientationMatrix2Rodrigues(om)
            # print(r)
            # and then the rotation angle
            # omega = 2 * np.arctan(np.linalg.norm(r)) * 180 / np.pi
            # todo: check if we can avoid computing the R vector
            cw = 0.5 * (om.trace() - 1)
            omega = np.arccos(cw)
            omegas.append(omega)
        index = np.argmin(omegas)
        if verbose:
            print(omegas)
            print('moving to FZ, index = %d' % index)
        return np.dot(syms[index], g)


class Lattice:
    '''
    The Lattice class to create one of the 14 Bravais lattices.

    This particular class has been partly inspired from the pymatgen
    project at https://github.com/materialsproject/pymatgen

    Any of the 7 lattice systems (each corresponding to one point group)
    can be easily created and manipulated.

    The lattice centering can be specified to form any of the 14 Bravais
    lattices:

     * Primitive (P): lattice points on the cell corners only (default);
     * Body (I): one additional lattice point at the center of the cell;
     * Face (F): one additional lattice point at the center of each of
       the faces of the cell;
     * Base (A, B or C): one additional lattice point at the center of
       each of one pair of the cell faces.

    ::

      a = 0.352 # FCC Nickel
      l = Lattice.face_centered_cubic(a)
      print(l.volume())

    Addditionnally the point-basis can be controlled to address non
    Bravais lattice cells. It is set to a single atoms at (0, 0, 0) by
    default so that each cell is a Bravais lattice but may be changed to
    something more complex to achieve HCP structure or Diamond structure
    for instance.
    '''

    def __init__(self, matrix, centering='P', symmetry=None):
        '''Create a crystal lattice (unit cell).

        Create a lattice from a 3x3 matrix.
        Each row in the matrix represents one lattice vector.
        '''
        m = np.array(matrix, dtype=np.float64).reshape((3, 3))
        lengths = np.sqrt(np.sum(m ** 2, axis=1))
        angles = np.zeros(3)
        for i in range(3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            angles[i] = dot(m[j], m[k]) / (lengths[j] * lengths[k])
        angles = np.arccos(angles) * 180. / pi
        self._angles = angles
        self._lengths = lengths
        self._matrix = m
        self._centering = centering
        self._symmetry = symmetry

    def __eq__(self, other):
        """Override the default Equals behavior.

        The equality of two Lattice objects is based on the equality of their angles, lengths, and centering.
        """
        if not isinstance(other, self.__class__):
            return False
        for i in range(3):
            if self._angles[i] != other._angles[i]:
                return False
            elif self._lengths[i] != other._lengths[i]:
                return False
        if self._centering != other._centering:
            return False
        if self._symmetry != other._symmetry:
            return False
        return True

    def __repr__(self):
        f = lambda x: "%0.1f" % x
        out = ["Lattice", " abc : " + " ".join(map(f, self._lengths)),
               " angles : " + " ".join(map(f, self._angles)),
               " volume : %0.4f" % self.volume(),
               " A : " + " ".join(map(f, self._matrix[0])),
               " B : " + " ".join(map(f, self._matrix[1])),
               " C : " + " ".join(map(f, self._matrix[2]))]
        return "\n".join(out)

    def reciprocal_lattice(self):
        '''Compute the reciprocal lattice.

        The reciprocal lattice defines a crystal in terms of vectors that
        are normal to a plane and whose lengths are the inverse of the
        interplanar spacing. This method computes the three reciprocal
        lattice vectors defined by:

        .. math::

         * a.a^* = 1
         * b.b^* = 1
         * c.c^* = 1
        '''
        [a, b, c] = self._matrix
        V = self.volume()
        astar = np.cross(b, c) / V
        bstar = np.cross(c, a) / V
        cstar = np.cross(a, b) / V
        return [astar, bstar, cstar]

    @property
    def matrix(self):
        """Returns a copy of matrix representing the Lattice."""
        return np.copy(self._matrix)

    def get_symmetry(self):
        """Returns the type of `Symmetry` of the Lattice."""
        return self._symmetry

    @staticmethod
    def symmetry(crystal_structure=Symmetry.cubic, use_miller_bravais=False):
        """Define the equivalent crystal symmetries.

        Those come from Randle & Engler, 2000. For instance in the cubic
        crystal struture, for instance there are 24 equivalent cube orientations.

        :param crystal_structure: an instance of the `Symmetry` class describing the crystal symmetry.
        :raise ValueError: if the given symmetry is not supported.
        :returns array: A numpy array of shape (n, 3, 3) where n is the \
        number of symmetries of the given crystal structure.
        """
        return crystal_structure.symmetry_operators(use_miller_bravais=use_miller_bravais)

    def get_lattice_parameters(self):
        """This function create a list of the independent lattice parameters depending on the symmetry.

        :return: a list of the lattice parameters.
        """
        sym = self.get_symmetry()
        (a, b, c) = self._lengths
        (alpha, beta, gamma) = self._angles
        # craft a list of the lattice parameters
        if sym is Symmetry.cubic:
            parameters = [a]
        elif sym in [Symmetry.hexagonal, Symmetry.trigonal, Symmetry.tetragonal]:
            parameters = [a, c]
        elif sym is Symmetry.orthorhombic:
            parameters = [a, b, c]
        elif sym is Symmetry.monoclinic:
            parameters = [a, b, c, alpha]
        else:
            parameters = [a, b, c, alpha, beta, gamma]
        return parameters

    def guess_symmetry(self):
        """Guess the lattice symmetry from the geometry."""
        (a, b, c) = self._lengths
        (alpha, beta, gamma) = self._angles
        return Lattice.guess_symmetry_from_parameters(a, b, c, alpha, beta, gamma)

    @staticmethod
    def guess_symmetry_from_parameters(a, b, c, alpha, beta, gamma):
        """Guess the lattice symmetry from the geometrical parameters."""
        if alpha == 90. and beta == 90. and gamma == 90:
            if a == b and a == c:
                return Symmetry.cubic
            elif a == b and a != c:
                return Symmetry.tetragonal
            else:
                return Symmetry.orthorhombic
        elif alpha == 90. and beta == 90. and gamma == 120 and a == b and a != c:
            return Symmetry.hexagonal
        elif a == b and a == c and alpha == beta and alpha == gamma:
            return Symmetry.trigonal
        elif a != b and a != c and beta == gamma and alpha != beta:
            return Symmetry.monoclinic
        else:
            return Symmetry.triclinic

    @staticmethod
    def from_cif(file_path):
        """
        Create a crystal Lattice using information contained in a given CIF
        file (Crystallographic Information Framework, a standard for
        information interchange in crystallography).

        Reference: S. R. Hall, F. H. Allen and I. D. Brown,
        The crystallographic information file (CIF): a new standard archive file for crystallography,
        Acta Crystallographica Section A, 47(6):655-685 (1991)
        doi = 10.1107/S010876739101067X

        .. note::

           Lattice constants are given in Angstrom in CIF files and so
           converted to nanometer.

        :param str file_path: The path to the CIF file representing the crystal structure.
        :returns: A `Lattice` instance corresponding to the given CIF file.
        """
        cf = CifFile.ReadCif(file_path)
        # crystal = eval('cf[\'%s\']' % symbol)
        crystal = cf.first_block()
        a = 0.1 * float(crystal['_cell_length_a'])
        b = 0.1 * float(crystal['_cell_length_b'])
        c = 0.1 * float(crystal['_cell_length_c'])
        alpha = float(crystal['_cell_angle_alpha'])
        beta = float(crystal['_cell_angle_beta'])
        gamma = float(crystal['_cell_angle_gamma'])
        try:
            symmetry = Symmetry.from_string(crystal['_symmetry_cell_setting'])
        except KeyError:
            symmetry = Lattice.guess_symmetry_from_parameters(a, b, c, alpha, beta, gamma)
        return Lattice.from_parameters(a, b, c, alpha, beta, gamma, symmetry=symmetry)

    @staticmethod
    def from_symbol(symbol):
        '''
        Create a crystal Lattice using information contained in a unit cell.

        *Parameters*

        **symbol**: The chemical symbol of the crystal (eg 'Al')

        *Returns*

        A `Lattice` instance corresponding to the given element.
        '''
        path = os.path.dirname(__file__)
        return Lattice.from_cif(os.path.join(path, 'cif', '%s.cif' % symbol))

    @staticmethod
    def cubic(a):
        '''
        Create a cubic Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter (a = b = c here)

        *Returns*

        A `Lattice` instance corresponding to a primitice cubic lattice.
        '''
        return Lattice([[a, 0.0, 0.0], [0.0, a, 0.0], [0.0, 0.0, a]], symmetry=Symmetry.cubic)

    @staticmethod
    def body_centered_cubic(a):
        '''
        Create a body centered cubic Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter (a = b = c here)

        *Returns*

        A `Lattice` instance corresponding to a body centered cubic
        lattice.
        '''
        return Lattice.from_parameters(a, a, a, 90, 90, 90, centering='I', symmetry=Symmetry.cubic)

    @staticmethod
    def face_centered_cubic(a):
        '''
        Create a face centered cubic Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter (a = b = c here)

        *Returns*

        A `Lattice` instance corresponding to a face centered cubic
        lattice.
        '''
        return Lattice.from_parameters(a, a, a, 90, 90, 90, centering='F', symmetry=Symmetry.cubic)

    @staticmethod
    def tetragonal(a, c):
        '''
        Create a tetragonal Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter

        **c**: third lattice length parameter (b = a here)

        *Returns*

        A `Lattice` instance corresponding to a primitive tetragonal
        lattice.
        '''
        return Lattice.from_parameters(a, a, c, 90, 90, 90, symmetry=Symmetry.tetragonal)

    @staticmethod
    def body_centered_tetragonal(a, c):
        '''
        Create a body centered tetragonal Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter

        **c**: third lattice length parameter (b = a here)

        *Returns*

        A `Lattice` instance corresponding to a body centered tetragonal
        lattice.
        '''
        return Lattice.from_parameters(a, a, c, 90, 90, 90, centering='I', symmetry=Symmetry.tetragonal)

    @staticmethod
    def orthorhombic(a, b, c):
        '''
        Create a tetragonal Lattice unit cell with 3 different length
        parameters a, b and c.
        '''
        return Lattice.from_parameters(a, b, c, 90, 90, 90, symmetry=Symmetry.orthorhombic)

    @staticmethod
    def base_centered_orthorhombic(a, b, c):
        '''
        Create a based centered orthorombic Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter

        **b**: second lattice length parameter

        **c**: third lattice length parameter

        *Returns*

        A `Lattice` instance corresponding to a based centered orthorombic
        lattice.
        '''
        return Lattice.from_parameters(a, b, c, 90, 90, 90, centering='C', symmetry=Symmetry.orthorhombic)

    @staticmethod
    def body_centered_orthorhombic(a, b, c):
        '''
        Create a body centered orthorombic Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter

        **b**: second lattice length parameter

        **c**: third lattice length parameter

        *Returns*

        A `Lattice` instance corresponding to a body centered orthorombic
        lattice.
        '''
        return Lattice.from_parameters(a, b, c, 90, 90, 90, centering='I', symmetry=Symmetry.orthorhombic)

    @staticmethod
    def face_centered_orthorhombic(a, b, c):
        '''
        Create a face centered orthorombic Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter

        **b**: second lattice length parameter

        **c**: third lattice length parameter

        *Returns*

        A `Lattice` instance corresponding to a face centered orthorombic
        lattice.
        '''
        return Lattice.from_parameters(a, b, c, 90, 90, 90, centering='F', symmetry=Symmetry.orthorhombic)

    @staticmethod
    def hexagonal(a, c):
        '''
        Create a hexagonal Lattice unit cell with length parameters a and c.
        '''
        return Lattice.from_parameters(a, a, c, 90, 90, 120, symmetry=Symmetry.hexagonal)

    @staticmethod
    def rhombohedral(a, alpha):
        '''
        Create a rhombohedral Lattice unit cell with one length
        parameter a and the angle alpha.
        '''
        return Lattice.from_parameters(a, a, a, alpha, alpha, alpha, symmetry=Symmetry.trigonal)

    @staticmethod
    def monoclinic(a, b, c, alpha):
        '''
        Create a monoclinic Lattice unit cell with 3 different length
        parameters a, b and c. The cell angle is given by alpha.
        The lattice centering id primitive ie. 'P'
        '''
        return Lattice.from_parameters(a, b, c, alpha, 90, 90, symmetry=Symmetry.monoclinic)

    @staticmethod
    def base_centered_monoclinic(a, b, c, alpha):
        '''
        Create a based centered monoclinic Lattice unit cell.

        *Parameters*

        **a**: first lattice length parameter

        **b**: second lattice length parameter

        **c**: third lattice length parameter

        **alpha**: first lattice angle parameter

        *Returns*

        A `Lattice` instance corresponding to a based centered monoclinic
        lattice.
        '''
        return Lattice.from_parameters(a, b, c, alpha, 90, 90, centering='C', symmetry=Symmetry.monoclinic)

    @staticmethod
    def triclinic(a, b, c, alpha, beta, gamma):
        '''
        Create a triclinic Lattice unit cell with 3 different length
        parameters a, b, c and three different cell angles alpha, beta
        and gamma.

        ..note::

           This method is here for the sake of completeness since one can
           create the triclinic cell directly using the `from_parameters`
           method.
        '''
        return Lattice.from_parameters(a, b, c, alpha, beta, gamma, symmetry=Symmetry.triclinic)

    @staticmethod
    def from_symmetry(symmetry, parameters):
        """Create a new lattice based on a type of symmetry and a list of lattice parameters.

        The type of symmetry should be an instance of `Symmetry` and the list of parameters should contain the
        appropriate number: 1 for cubic, 2 for hexagonal, tetragonal or trigonal, 3 for orthorhombic, 4 for monoclinic
        or 6 for triclinic.

        :param symmetry: an instance of `Symmetry`.
        :param list parameters: a list of the lattice parameters.
        :return: the newly created `Lattice` instance.
        """
        if symmetry is Symmetry.cubic:
            if len(parameters) != 1:
                raise(ValueError('The number of parameters for %s symmetry should be 1, got %d' % (symmetry, len(parameters))))
            return Lattice.cubic(parameters[0])
        elif symmetry in [Symmetry.hexagonal, Symmetry.trigonal]:
            if len(parameters) != 2:
                raise(ValueError('The number of parameters for %s symmetry should be 2, got %d' % (symmetry, len(parameters))))
            return Lattice.hexagonal(parameters[0], parameters[1])
        elif symmetry is Symmetry.orthorhombic:
            if len(parameters) != 3:
                raise(ValueError('The number of parameters for %s symmetry should be 3, got %d' % (symmetry, len(parameters))))
            return Lattice.orthorhombic(parameters[0], parameters[1], parameters[2])
        elif symmetry is Symmetry.tetragonal:
            if len(parameters) != 2:
                raise(ValueError('The number of parameters for %s symmetry should be 2, got %d' % (symmetry, len(parameters))))
            return Lattice.tetragonal(parameters[0], parameters[1])
        elif symmetry is Symmetry.monoclinic:
            if len(parameters) != 4:
                raise(ValueError('The number of parameters for %s symmetry should be 4, got %d' % (symmetry, len(parameters))))
            return Lattice.monoclinic(parameters[0], parameters[1], parameters[2], parameters[3])
        else:
            if len(parameters) != 6:
                raise(ValueError('The number of parameters for triclinic symmetry should be 6, got %d' % len(parameters)))
            return Lattice.triclinic(*parameters)

    @staticmethod
    def from_parameters(a, b, c, alpha, beta, gamma, x_aligned_with_a=True, centering='P', symmetry=Symmetry.triclinic):
        """
        Create a Lattice using unit cell lengths and angles (in degrees).
        The lattice centering can also be specified (among 'P', 'I', 'F',
        'A', 'B' or 'C').

        :param float a: first lattice length parameter.
        :param float b: second lattice length parameter.
        :param float c: third lattice length parameter.
        :param float alpha: first lattice angle parameter.
        :param float beta: second lattice angle parameter.
        :param float gamma: third lattice angle parameter.
        :param bool x_aligned_with_a: flag to control the convention used to define the Cartesian frame.
        :param str centering: lattice centering ('P' by default) passed to the `Lattice` class.
        :param symmetry: a `Symmetry` instance to be passed to the lattice.
        :return: A `Lattice` instance with the specified lattice parameters and centering.
        """
        alpha_r = radians(alpha)
        beta_r = radians(beta)
        gamma_r = radians(gamma)
        if x_aligned_with_a:  # first lattice vector (a) is aligned with X
            vector_a = a * np.array([1, 0, 0])
            vector_b = b * np.array([np.cos(gamma_r), np.sin(gamma_r), 0])
            c1 = c * np.cos(beta_r)
            c2 = c * (np.cos(alpha_r) - np.cos(gamma_r) * np.cos(beta_r)) / np.sin(gamma_r)
            vector_c = np.array([c1, c2, np.sqrt(c ** 2 - c1 ** 2 - c2 ** 2)])
        else:  # third lattice vector (c) is aligned with Z
            cos_gamma_star = (np.cos(alpha_r) * np.cos(beta_r) - np.cos(gamma_r)) / (np.sin(alpha_r) * np.sin(beta_r))
            sin_gamma_star = np.sqrt(1 - cos_gamma_star ** 2)
            vector_a = [a * np.sin(beta_r), 0.0, a * np.cos(beta_r)]
            vector_b = [-b * np.sin(alpha_r) * cos_gamma_star, b * np.sin(alpha_r) * sin_gamma_star, b * np.cos(alpha_r)]
            vector_c = [0.0, 0.0, float(c)]
        return Lattice([vector_a, vector_b, vector_c], centering=centering, symmetry=symmetry)

    def volume(self):
        """Compute the volume of the unit cell."""
        m = self._matrix
        return abs(np.dot(np.cross(m[0], m[1]), m[2]))

    def get_hkl_family(self, hkl):
        """Get a list of the hkl planes composing the given family for
        this crystal lattice.

        *Parameters*

        **hkl**: miller indices of the requested family

        *Returns*

        A list of the hkl planes in the given family.
        """
        planes = HklPlane.get_family(hkl, lattice=self, crystal_structure=self._symmetry)
        return planes


class SlipSystem:
    '''A class to represent a crystallographic slip system.

    A slip system is composed of a slip plane (most widely spaced planes
    in the crystal) and a slip direction (highest linear density of atoms
    in the crystal).
    '''

    def __init__(self, plane, direction):
        '''Create a new slip system object with the given slip plane and
        slip direction.
        '''
        self._plane = plane
        self._direction = direction

    def __repr__(self):
        out = '(%d%d%d)' % self._plane.miller_indices()
        out += '[%d%d%d]' % self._direction.miller_indices()
        return out

    def get_slip_plane(self):
        return self._plane

    def get_slip_direction(self):
        return self._direction

    @staticmethod
    def from_indices(plane_indices, direction_indices, lattice=None):
        '''A static method to create a slip system from the indices of the plane and the direction.
        '''
        plane = HklPlane(plane_indices[0], plane_indices[1], plane_indices[2], lattice)
        direction = HklDirection(direction_indices[0], direction_indices[1], direction_indices[2], lattice)
        return SlipSystem(plane, direction)

    @staticmethod
    def get_slip_systems(plane_type='111'):
        '''A static method to get all slip systems for a given hkl plane family.

        :params str plane_type: a string of the 3 miller indices of the crystallographic plane family.
        :returns list: a list of :py:class:`~pymicro.crystal.lattice.SlipSystem`.

        .. warning::

          only working for 111 and 112 planes...
        '''
        slip_systems = []
        if plane_type == '001':
            slip_systems.append(SlipSystem(HklPlane(0, 0, 1), HklDirection(-1, 1, 0)))  # E5
            slip_systems.append(SlipSystem(HklPlane(0, 0, 1), HklDirection(1, 1, 0)))  # E6
            slip_systems.append(SlipSystem(HklPlane(1, 0, 0), HklDirection(0, 1, 1)))  # F1
            slip_systems.append(SlipSystem(HklPlane(1, 0, 0), HklDirection(0, -1, 1)))  # F2
            slip_systems.append(SlipSystem(HklPlane(0, 1, 0), HklDirection(-1, 0, 1)))  # G4
            slip_systems.append(SlipSystem(HklPlane(0, 1, 0), HklDirection(1, 0, 1)))  # G3
        elif plane_type == '111':
            slip_systems.append(SlipSystem(HklPlane(1, 1, 1), HklDirection(-1, 0, 1)))  # Bd
            slip_systems.append(SlipSystem(HklPlane(1, 1, 1), HklDirection(0, -1, 1)))  # Ba
            slip_systems.append(SlipSystem(HklPlane(1, 1, 1), HklDirection(-1, 1, 0)))  # Bc
            slip_systems.append(SlipSystem(HklPlane(1, -1, 1), HklDirection(-1, 0, 1)))  # Db
            slip_systems.append(SlipSystem(HklPlane(1, -1, 1), HklDirection(0, 1, 1)))  # Dc
            slip_systems.append(SlipSystem(HklPlane(1, -1, 1), HklDirection(1, 1, 0)))  # Da
            slip_systems.append(SlipSystem(HklPlane(-1, 1, 1), HklDirection(0, -1, 1)))  # Ab
            slip_systems.append(SlipSystem(HklPlane(-1, 1, 1), HklDirection(1, 1, 0)))  # Ad
            slip_systems.append(SlipSystem(HklPlane(-1, 1, 1), HklDirection(1, 0, 1)))  # Ac
            slip_systems.append(SlipSystem(HklPlane(1, 1, -1), HklDirection(-1, 1, 0)))  # Cb
            slip_systems.append(SlipSystem(HklPlane(1, 1, -1), HklDirection(1, 0, 1)))  # Ca
            slip_systems.append(SlipSystem(HklPlane(1, 1, -1), HklDirection(0, 1, 1)))  # Cd
        elif plane_type == '112':
            slip_systems.append(SlipSystem(HklPlane(1, 1, 2), HklDirection(1, 1, -1)))
            slip_systems.append(SlipSystem(HklPlane(-1, 1, 2), HklDirection(1, -1, 1)))
            slip_systems.append(SlipSystem(HklPlane(1, -1, 2), HklDirection(-1, 1, 1)))
            slip_systems.append(SlipSystem(HklPlane(1, 1, -2), HklDirection(1, 1, 1)))
            slip_systems.append(SlipSystem(HklPlane(1, 2, 1), HklDirection(1, -1, 1)))
            slip_systems.append(SlipSystem(HklPlane(-1, 2, 1), HklDirection(1, 1, -1)))
            slip_systems.append(SlipSystem(HklPlane(1, -2, 1), HklDirection(1, 1, 1)))
            slip_systems.append(SlipSystem(HklPlane(1, 2, -1), HklDirection(-1, 1, 1)))
            slip_systems.append(SlipSystem(HklPlane(2, 1, 1), HklDirection(-1, 1, 1)))
            slip_systems.append(SlipSystem(HklPlane(-2, 1, 1), HklDirection(1, 1, 1)))
            slip_systems.append(SlipSystem(HklPlane(2, -1, 1), HklDirection(1, 1, -1)))
            slip_systems.append(SlipSystem(HklPlane(2, 1, -1), HklDirection(1, -1, 1)))
        else:
            print('warning only 001, 111 or 112 slip planes supported for the moment!')
        return slip_systems


class HklObject:
    def __init__(self, h, k, l, lattice=None):
        '''Create a new hkl object with the given Miller indices and
           crystal lattice.
        '''
        if lattice == None:
            lattice = Lattice.cubic(1.0)
        self._lattice = lattice
        self._h = h
        self._k = k
        self._l = l

    @property
    def lattice(self):
        return self._lattice

    def set_lattice(self, lattice):
        """Assign a new `Lattice` to this instance.

        :param lattice: the new crystal lattice.
        """
        self._lattice = lattice

    @property
    def h(self):
        return self._h

    @property
    def k(self):
        return self._k

    @property
    def l(self):
        return self._l

    def miller_indices(self):
        '''
        Returns an immutable tuple of the plane Miller indices.
        '''
        return (self._h, self._k, self._l)

    @staticmethod
    def skip_higher_order(hkl_list, keep_friedel_pair=False, verbose=False):
        """Create a copy of a list of some hkl object retaining only the first order.

        :param list hkl_list: The list of `HklObject`.
        :param bool keep_friedel_pair: flag to keep order -1 in the list. 
        :param bool verbose: activate verbose mode.
        :returns list: A new list of :py:class:`~pymicro.crystal.lattice.HklObject` without any multiple reflection.
        """
        # create array with all the miller indices
        hkl_array = np.empty((len(hkl_list), 3), dtype=int)
        for i in range(len(hkl_list)):
            hkl = hkl_list[i]
            hkl_array[i] = np.array(hkl.miller_indices())
        # first start by ordering the HklObjects by ascending miller indices sum
        hkl_sum = np.sum(np.abs(hkl_array), axis=1)
        hkl_sum_sort = np.argsort(hkl_sum)
        first_order_list = [hkl_list[hkl_sum_sort[0]]]
        for i in range(1, len(hkl_sum_sort)):
            hkl_next = hkl_sum_sort[i]
            hkl = hkl_list[hkl_next]
            (h, k, l) = hkl.miller_indices()
            if verbose:
                print('trying hkl object (%d, %d, %d)' % (h, k, l))
            lower = False
            # check if a hkl object already exist in the list
            for uvw in first_order_list:
                # try to assess the multiple from the sum of miller indices
                (u, v, w) = uvw.miller_indices()
                if verbose:
                    print('looking at: (%d, %d, %d)' % (u, v, w))
                n = hkl_sum[hkl_next] / np.sum(np.abs(np.array((u, v, w))), axis=0)
                for order in [-n, n]:
                    if keep_friedel_pair and order == -1:
                        if verbose:
                            print('keeping Friedel pair reflexion: (%d, %d, %d) with n=%d' % (u, v, w, order))
                        continue
                    if (u * order == h) and (v * order == k) and (w * order) == l:
                        if verbose:
                            print('lower order reflexion was found: (%d, %d, %d) with n=%d' % (u, v, w, order))
                        lower = True
                        break
            # if no lower order reflexion was found, add the hkl object to the list
            if not lower:
                if verbose:
                    print('adding hkl object (%d, %d, %d) to the list' % (h, k, l))
                first_order_list.append(hkl)
        return first_order_list


class HklDirection(HklObject):
    def __repr__(self):
        f = lambda x: "%0.3f" % x
        out = ['HKL Direction',
               ' Miller indices:',
               ' h : ' + str(self._h),
               ' k : ' + str(self._k),
               ' l : ' + str(self._l),
               ' crystal lattice : ' + str(self._lattice)]
        return '\n'.join(out)

    def direction(self):
        '''Returns a normalized vector, expressed in the cartesian
        coordinate system, corresponding to this crystallographic direction.
        '''
        (h, k, l) = self.miller_indices()
        M = self._lattice.matrix.T  # the columns of M are the a, b, c vector in the cartesian coordinate system
        l_vect = M.dot(np.array([h, k, l]))
        return l_vect / np.linalg.norm(l_vect)

    def angle_with_direction(self, hkl):
        '''Computes the angle between this crystallographic direction and
        the given direction (in radian).'''
        return np.arccos(np.dot(self.direction(), hkl.direction()))

    @staticmethod
    def angle_between_directions(hkl1, hkl2, lattice=None):
        '''Computes the angle between two crystallographic directions (in radian).

        :param tuple hkl1: The triplet of the miller indices of the first direction.
        :param tuple hkl2: The triplet of the miller indices of the second direction.
        :param Lattice lattice: The crystal lattice, will default to cubic if not specified.

        :returns float: The angle in radian.
        '''
        d1 = HklDirection(*hkl1, lattice=lattice)
        d2 = HklDirection(*hkl2, lattice=lattice)
        return d1.angle_with_direction(d2)

    @staticmethod
    def three_to_four_indices(u, v, w):
        """Convert from Miller indices to Miller-Bravais indices. this is used for hexagonal crystal lattice."""
        return (2 * u - v) / 3., (2 * v - u) / 3., -(u + v) / 3., w

    @staticmethod
    def four_to_three_indices(U, V, T, W):
        """Convert from Miller-Bravais indices to Miller indices. this is used for hexagonal crystal lattice."""
        import fractions
        u, v, w = U - T, V - T, W
        gcd = functools.reduce(fractions.gcd, (u, v, w))
        return u / gcd, v / gcd, w / gcd

    @staticmethod
    def angle_between_4indices_directions(hkil1, hkil2, ac):
        """Computes the angle between two crystallographic directions in a hexagonal lattice.

        The solution was derived by F. Frank in:
        On Miller - Bravais indices and four dimensional vectors. Acta Cryst. 18, 862-866 (1965)

        :param tuple hkil1: The quartet of the indices of the first direction.
        :param tuple hkil2: The quartet of the indices of the second direction.
        :param tuple ac: the lattice parameters of the hexagonal structure in the form (a, c).
        :returns float: The angle in radian.
        """
        h1, k1, i1, l1 = hkil1
        h2, k2, i2, l2 = hkil2
        a, c = ac
        lambda_square = 2. / 3 * (c / a) ** 2
        value = (h1 * h2 + k1 * k2 + i1 * i2 + lambda_square * l1 * l2) / \
                (np.sqrt(h1 ** 2 + k1 ** 2 + i1 ** 2 + lambda_square * l1 ** 2) *
                 np.sqrt(h2 ** 2 + k2 ** 2 + i2 ** 2 + lambda_square * l2 ** 2))
        return np.arccos(value)

    def find_planes_in_zone(self, max_miller=5):
        '''
        This method finds the hkl planes in zone with the crystallographic
        direction. If (u,v,w) denotes the zone axis, this means finding all
        hkl planes which verify :math:`h.u + k.v + l.w = 0`.

        :param max_miller: The maximum miller index to limt the search`
        :returns list: A list of :py:class:`~pymicro.crystal.lattice.HklPlane` objects \
        describing all the planes in zone with the direction.
        '''
        (u, v, w) = self.miller_indices()
        indices = range(-max_miller, max_miller + 1)
        hklplanes_in_zone = []
        for h in indices:
            for k in indices:
                for l in indices:
                    if h == k == l == 0:  # skip (0, 0, 0)
                        continue
                    if np.dot(np.array([h, k, l]), np.array([u, v, w])) == 0:
                        hklplanes_in_zone.append(HklPlane(h, k, l, self._lattice))
        return hklplanes_in_zone


class HklPlane(HklObject):
    '''
    This class define crystallographic planes using Miller indices.

    A plane can be create by speficying its Miller indices and the
    crystal lattice (default is cubic with lattice parameter of 1.0).

    ::

      a = 0.405 # FCC Aluminium
      l = Lattice.cubic(a)
      p = HklPlane(1, 1, 1, lattice=l)
      print(p)
      print(p.scattering_vector())
      print(p.interplanar_spacing())

    .. note::

      Miller indices are defined in terms of the inverse of the intercept
      of the plane on the three crystal axes a, b, and c.
    '''

    def __eq__(self, other):
        """Override the default Equals behavior.

        The equality of two HklObjects is based on the equality of their miller indices.
        """
        if isinstance(other, self.__class__):
            return self._h == other._h and self._k == other._k and \
                   self._l == other._l and self._lattice == other._lattice
        return False

    def __ne__(self, other):
        """Define a non-equality test"""
        return not self.__eq__(other)

    def normal(self):
        '''Returns the unit vector normal to the plane.

        We use of the repiprocal lattice to compute the normal to the plane
        and return a normalised vector.
        '''
        n = self.scattering_vector()
        return n / np.linalg.norm(n)

    def scattering_vector(self):
        '''Calculate the scattering vector of this `HklPlane`.

        The scattering vector (or reciprocal lattice vector) is normal to
        this `HklPlane` and its length is equal to the inverse of the
        interplanar spacing. In the cartesian coordinate system of the
        crystal, it is given by:

        ..math

          G_c = h.a^* + k.b^* + l.c^*

        :returns: a numpy vector expressed in the cartesian coordinate system of the crystal.
        '''
        [astar, bstar, cstar] = self._lattice.reciprocal_lattice()
        (h, k, l) = self.miller_indices()
        # express (h, k, l) in the cartesian crystal CS
        Gc = h * astar + k * bstar + l * cstar
        return Gc

    def __repr__(self):
        f = lambda x: "%0.3f" % x
        out = ['HKL Plane',
               ' Miller indices:',
               ' h : ' + str(self._h),
               ' k : ' + str(self._k),
               ' l : ' + str(self._l),
               ' plane normal : ' + str(self.normal()),
               ' crystal lattice : ' + str(self._lattice)]
        return '\n'.join(out)

    def friedel_pair(self):
        """Create the Friedel pair of the HklPlane."""
        (h, k, l) = self.miller_indices()
        pair = HklPlane(-h, -k, -l, self._lattice)
        return pair

    def interplanar_spacing(self):
        '''
        Compute the interplanar spacing.
        For cubic lattice, it is:

        .. math::

           d = a / \sqrt{h^2 + k^2 + l^2}

        The general formula comes from 'Introduction to Crystallography'
        p. 68 by Donald E. Sands.
        '''
        (a, b, c) = self._lattice._lengths
        (h, k, l) = self.miller_indices()
        (alpha, beta, gamma) = radians(self._lattice._angles)
        # d = a / np.sqrt(h**2 + k**2 + l**2) # for cubic structure only
        d = self._lattice.volume() / np.sqrt(h ** 2 * b ** 2 * c ** 2 * np.sin(alpha) ** 2 + \
                                             k ** 2 * a ** 2 * c ** 2 * np.sin(
                                                 beta) ** 2 + l ** 2 * a ** 2 * b ** 2 * np.sin(gamma) ** 2 + \
                                             2 * h * l * a * b ** 2 * c * (
                                                 np.cos(alpha) * np.cos(gamma) - np.cos(beta)) + \
                                             2 * h * k * a * b * c ** 2 * (
                                                 np.cos(alpha) * np.cos(beta) - np.cos(gamma)) + \
                                             2 * k * l * a ** 2 * b * c * (
                                                 np.cos(beta) * np.cos(gamma) - np.cos(alpha)))
        return d

    def bragg_angle(self, lambda_keV, verbose=False):
        '''Compute the Bragg angle for this `HklPlane` at the given energy.

        .. note::

          For this calculation to work properly, the lattice spacing needs
          to be in nm units.
        '''
        d = self.interplanar_spacing()
        lambda_nm = 1.2398 / lambda_keV
        theta = np.arcsin(lambda_nm / (2 * d))
        if verbose:
            theta_deg = 180 * theta / np.pi
            (h, k, l) = self.miller_indices()
            print('\nBragg angle for %d%d%d at %.1f keV is %.1f deg\n' % (h, k, l, lambda_keV, theta_deg))
        return theta

    @staticmethod
    def four_to_three_indices(U, V, T, W):
        """Convert four to three index representation of a slip plane (used for hexagonal crystal lattice)."""
        #return (6 * h / 5. - 3 * k / 5., 3 * h / 5. + 6 * k / 5., l)
        return U, V, W

    @staticmethod
    def three_to_four_indices(u, v, w):
        """Convert three to four index representation of a slip plane (used for hexagonal crystal lattice)."""
        return u, v, -(u + v), w

    def is_in_list(self, hkl_planes, friedel_pair=False):
        """Check if the hkl plane is in the given list.

        By default this relies on the built in in test from the list type which in turn calls in the __eq__ method.
        This means it will return True if a plane with the exact same miller indices (and same lattice) is in the list.
        Turning on the friedel_pair flag will allow to test also the Friedel pair (-h, -k, -l) and return True if it is
        in the list.
        For instance (0,0,1) and (0,0,-1) are in general considered as the same lattice plane.
        """
        if not friedel_pair:
            return self in hkl_planes
        else:
            return self in hkl_planes or self.friedel_pair() in hkl_planes

    @staticmethod
    def is_same_family(hkl1, hkl2, crystal_structure=Symmetry.cubic):
        """Static method to test if both lattice planes belongs to the same family.
        
        A family {hkl} is composed by all planes that are equivalent to (hkl) using the symmetry of the lattice. 
        The lattice assoiated with `hkl2`is not taken into account here.
        """
        return hkl1.is_in_list(HklPlane.get_family(hkl2.miller_indices(), lattice=hkl1._lattice,
                                                   crystal_structure=crystal_structure))

    @staticmethod
    def get_family(hkl, lattice=None, include_friedel_pairs=False, crystal_structure=Symmetry.cubic):
        """Static method to obtain a list of the different crystallographic
        planes in a particular family.

        :param str hkl: a sequence of 3 (4 for hexagonal) numbers corresponding to the miller indices.
        :param Lattice lattice: The reference crystal lattice (default None).
        :param bool include_friedel_pairs: Flag to include the Friedel pairs in the list (False by default).
        :param str crystal_structure: A string descibing the crystal structure (cubic by default). 
        :raise ValueError: if the given string does not correspond to a supported family.
        :returns list: a list of the :py:class:`~pymicro.crystal.lattice.HklPlane` in the given hkl family.

        .. note::

          The method account for the lattice symmetry to create a list of equivalent lattice plane from the point 
          of view of the point group symmetry. A flag can be used to include or not the Friedel pairs. If not, the 
          family is contstructed using the miller indices limited the number of minus signs. For instance  (1,0,0) 
          will be in the list and not (-1,0,0).
        """
        if not (len(hkl) == 3 or (len(hkl) == 4 and crystal_structure == Symmetry.hexagonal)):
            raise ValueError('warning, family not supported: {}'.format(hkl))
        # handle hexagonal case
        if len(hkl) == 4:
            h = int(hkl[0])
            k = int(hkl[1])
            i = int(hkl[2])
            l = int(hkl[3])
            (h, k, l) = HklPlane.four_to_three_indices(h, k, i, l)  # useless as it just drops i
        else:  # 3 indices
            h = int(hkl[0])
            k = int(hkl[1])
            l = int(hkl[2])
            if crystal_structure == Symmetry.hexagonal:
                i = -(h + k)
        family = []
        # construct lattice plane family from the symmetry operators
        if crystal_structure == Symmetry.hexagonal:
          syms = Lattice.symmetry(crystal_structure, use_miller_bravais=True)  # use 4x4 symmetry operators
        else:
          syms = Lattice.symmetry(crystal_structure)
        for sym in syms:
            if crystal_structure == Symmetry.hexagonal:
                n_sym = np.dot(sym, np.array([h, k, i, l]))
                n_sym = HklPlane.four_to_three_indices(*n_sym)
            else:  # 3 indices
                n_sym = np.dot(sym, np.array([h, k, l]))
            hkl_sym = HklPlane(*n_sym, lattice=lattice)
            if not hkl_sym.is_in_list(family, friedel_pair=True):
                family.append(hkl_sym)
            if include_friedel_pairs:
                hkl_sym = HklPlane(-n_sym[0], -n_sym[1], -n_sym[2], lattice=lattice)
                if not hkl_sym.is_in_list(family, friedel_pair=False):
                    family.append(hkl_sym)
        if not include_friedel_pairs:
            # for each hkl plane chose between (h, k, l) and (-h, -k, -l) to have the less minus signs
            for i in range(len(family)):
                hkl = family[i]
                (h, k, l) = hkl.miller_indices()
                if np.where(np.array([h, k, l]) < 0)[0].size > 0 and np.where(np.array([h, k, l]) <= 0)[0].size >= 2:
                    family[i] = hkl.friedel_pair()
                    #print('replacing plane (%d%d%d) by its pair: (%d%d%d)' % (h, k, l, -h, -k, -l))


        '''
        if not len(hkl) == 3:
            raise ValueError('warning, family not supported: %s' % hkl)
        family = []
        hkl_list = list(hkl)
        hkl_list.sort()  # miller indices are now sorted by increasing order
        h = int(hkl[0])
        k = int(hkl[1])
        l = int(hkl[2])
        if hkl == '001' or hkl == '010' or hkl == '100':
            family.append(HklPlane(1, 0, 0, lattice))
            family.append(HklPlane(0, 1, 0, lattice))
            family.append(HklPlane(0, 0, 1, lattice))
        elif hkl in ['011', '101', '110']:
            family.append(HklPlane(1, 1, 0, lattice))
            family.append(HklPlane(-1, 1, 0, lattice))
            family.append(HklPlane(1, 0, 1, lattice))
            family.append(HklPlane(-1, 0, 1, lattice))
            family.append(HklPlane(0, 1, 1, lattice))
            family.append(HklPlane(0, -1, 1, lattice))
        elif hkl == '111':
            family.append(HklPlane(1, 1, 1, lattice))
            family.append(HklPlane(-1, 1, 1, lattice))
            family.append(HklPlane(1, -1, 1, lattice))
            family.append(HklPlane(1, 1, -1, lattice))
        elif hkl in ['112', '121', '211']:
            family.append(HklPlane(1, 1, 2, lattice))
            family.append(HklPlane(-1, 1, 2, lattice))
            family.append(HklPlane(1, -1, 2, lattice))
            family.append(HklPlane(1, 1, -2, lattice))
            family.append(HklPlane(1, 2, 1, lattice))
            family.append(HklPlane(-1, 2, 1, lattice))
            family.append(HklPlane(1, -2, 1, lattice))
            family.append(HklPlane(1, 2, -1, lattice))
            family.append(HklPlane(2, 1, 1, lattice))
            family.append(HklPlane(-2, 1, 1, lattice))
            family.append(HklPlane(2, -1, 1, lattice))
            family.append(HklPlane(2, 1, -1, lattice))
        elif hkl in ['113', '131', '311']:
            family.append(HklPlane(1, 1, 3, lattice))
            family.append(HklPlane(-1, 1, 3, lattice))
            family.append(HklPlane(1, -1, 3, lattice))
            family.append(HklPlane(1, 1, -3, lattice))
            family.append(HklPlane(1, 3, 1, lattice))
            family.append(HklPlane(-1, 3, 1, lattice))
            family.append(HklPlane(1, -3, 1, lattice))
            family.append(HklPlane(1, 3, -1, lattice))
            family.append(HklPlane(3, 1, 1, lattice))
            family.append(HklPlane(-3, 1, 1, lattice))
            family.append(HklPlane(3, -1, 1, lattice))
            family.append(HklPlane(3, 1, -1, lattice))
        elif hkl in ['331', '313', '133']:
            family.append(HklPlane(3, 3, 1, lattice))
            family.append(HklPlane(-3, 3, 1, lattice))
            family.append(HklPlane(3, -3, 1, lattice))
            family.append(HklPlane(3, 3, -1, lattice))
            family.append(HklPlane(3, 1, 3, lattice))
            family.append(HklPlane(-3, 1, 3, lattice))
            family.append(HklPlane(3, -1, 3, lattice))
            family.append(HklPlane(3, 1, -3, lattice))
            family.append(HklPlane(1, 3, 3, lattice))
            family.append(HklPlane(-1, 3, 3, lattice))
            family.append(HklPlane(1, -3, 3, lattice))
            family.append(HklPlane(1, 3, -3, lattice))
        elif hkl in ['662', '626', '266']:
            family.append(HklPlane(6, 6, 2, lattice))
            family.append(HklPlane(-6, 6, 2, lattice))
            family.append(HklPlane(6, -6, 2, lattice))
            family.append(HklPlane(6, 6, -2, lattice))
            family.append(HklPlane(6, 2, 6, lattice))
            family.append(HklPlane(-6, 2, 6, lattice))
            family.append(HklPlane(6, -2, 6, lattice))
            family.append(HklPlane(6, 2, -6, lattice))
            family.append(HklPlane(2, 6, 6, lattice))
            family.append(HklPlane(-2, 6, 6, lattice))
            family.append(HklPlane(2, -6, 6, lattice))
            family.append(HklPlane(2, 6, -6, lattice))
        elif hkl == '002':
            family.append(HklPlane(2, 0, 0, lattice))
            family.append(HklPlane(0, 2, 0, lattice))
            family.append(HklPlane(0, 0, 2, lattice))
        elif hkl == '022':
            family.append(HklPlane(2, 2, 0, lattice))
            family.append(HklPlane(-2, 2, 0, lattice))
            family.append(HklPlane(2, 0, 2, lattice))
            family.append(HklPlane(-2, 0, 2, lattice))
            family.append(HklPlane(0, 2, 2, lattice))
            family.append(HklPlane(0, -2, 2, lattice))
        elif hkl == '123':
            family.append(HklPlane(1, 2, 3, lattice))
            family.append(HklPlane(-1, 2, 3, lattice))
            family.append(HklPlane(1, -2, 3, lattice))
            family.append(HklPlane(1, 2, -3, lattice))
            family.append(HklPlane(3, 1, 2, lattice))
            family.append(HklPlane(-3, 1, 2, lattice))
            family.append(HklPlane(3, -1, 2, lattice))
            family.append(HklPlane(3, 1, -2, lattice))
            family.append(HklPlane(2, 3, 1, lattice))
            family.append(HklPlane(-2, 3, 1, lattice))
            family.append(HklPlane(2, -3, 1, lattice))
            family.append(HklPlane(2, 3, -1, lattice))
            family.append(HklPlane(1, 3, 2, lattice))
            family.append(HklPlane(-1, 3, 2, lattice))
            family.append(HklPlane(1, -3, 2, lattice))
            family.append(HklPlane(1, 3, -2, lattice))
            family.append(HklPlane(2, 1, 3, lattice))
            family.append(HklPlane(-2, 1, 3, lattice))
            family.append(HklPlane(2, -1, 3, lattice))
            family.append(HklPlane(2, 1, -3, lattice))
            family.append(HklPlane(3, 2, 1, lattice))
            family.append(HklPlane(-3, 2, 1, lattice))
            family.append(HklPlane(3, -2, 1, lattice))
            family.append(HklPlane(3, 2, -1, lattice))
        elif len(np.unique(list(hkl))) == 1 and not '0' in hkl:
            # hhh planes -> 4 planes
            family.append(HklPlane(h, h, h, lattice))
            family.append(HklPlane(-h, h, h, lattice))
            family.append(HklPlane(h, -h, h, lattice))
            family.append(HklPlane(h, h, -h, lattice))
        elif len(np.unique(hkl_list)) == 2 and hkl_list[0] == hkl_list[1] and not '0' in hkl:
            # 2 different ints, hhl type -> 12 planes
            family.append(HklPlane(h, h, l, lattice))
            family.append(HklPlane(-h, h, l, lattice))
            family.append(HklPlane(h, -h, l, lattice))
            family.append(HklPlane(h, h, -l, lattice))
            family.append(HklPlane(h, l, h, lattice))
            family.append(HklPlane(-h, l, h, lattice))
            family.append(HklPlane(h, -l, h, lattice))
            family.append(HklPlane(h, l, -h, lattice))
            family.append(HklPlane(l, h, h, lattice))
            family.append(HklPlane(-l, h, h, lattice))
            family.append(HklPlane(l, -h, h, lattice))
            family.append(HklPlane(l, h, -h, lattice))
        elif len(np.unique(hkl_list)) == 2 and hkl_list[1] == hkl_list[2] and not '0' in hkl:
            # 2 different ints, hkk type -> 12 planes
            family.append(HklPlane(k, k, h, lattice))
            family.append(HklPlane(-k, k, h, lattice))
            family.append(HklPlane(k, -k, h, lattice))
            family.append(HklPlane(k, k, -h, lattice))
            family.append(HklPlane(k, h, k, lattice))
            family.append(HklPlane(-k, h, k, lattice))
            family.append(HklPlane(k, -h, k, lattice))
            family.append(HklPlane(k, h, -k, lattice))
            family.append(HklPlane(h, k, k, lattice))
            family.append(HklPlane(-h, k, k, lattice))
            family.append(HklPlane(h, -k, k, lattice))
            family.append(HklPlane(h, k, -k, lattice))
        elif len(np.unique(list(hkl))) == 3 and not '0' in hkl:
            # 3 different ints, all nonzeros -> 24 planes
            family.append(HklPlane(h, k, l, lattice))
            family.append(HklPlane(-h, k, l, lattice))
            family.append(HklPlane(h, -k, l, lattice))
            family.append(HklPlane(h, k, -l, lattice))
            family.append(HklPlane(l, h, k, lattice))
            family.append(HklPlane(-l, h, k, lattice))
            family.append(HklPlane(l, -h, k, lattice))
            family.append(HklPlane(l, h, -k, lattice))
            family.append(HklPlane(k, l, h, lattice))
            family.append(HklPlane(-k, l, h, lattice))
            family.append(HklPlane(k, -l, h, lattice))
            family.append(HklPlane(k, l, -h, lattice))
            family.append(HklPlane(h, l, k, lattice))
            family.append(HklPlane(-h, l, k, lattice))
            family.append(HklPlane(h, -l, k, lattice))
            family.append(HklPlane(h, l, -k, lattice))
            family.append(HklPlane(k, h, l, lattice))
            family.append(HklPlane(-k, h, l, lattice))
            family.append(HklPlane(k, -h, l, lattice))
            family.append(HklPlane(k, h, -l, lattice))
            family.append(HklPlane(l, k, h, lattice))
            family.append(HklPlane(-l, k, h, lattice))
            family.append(HklPlane(l, -k, h, lattice))
            family.append(HklPlane(l, k, -h, lattice))
        else:
            raise ValueError('warning, family not supported: %s' % hkl)
        '''
        return family

    def multiplicity(self, symmetry=Symmetry.cubic):
        """compute the general multiplicity for this `HklPlane` and the given `Symmetry`.

        :param Symmetry symmetry: The crystal symmetry to take into account.
        :return: the number of equivalent planes in the family.
        """
        return len(HklPlane.get_family(self.miller_indices(), include_friedel_pairs=True, crystal_structure=symmetry))

    def slip_trace(self, orientation, n_int=np.array([0, 0, 1]), view_up=np.array([0, 1, 0]), trace_size=100, verbose=False):
        """
        Compute the intersection of the lattice plane with a particular plane defined by its normal.

        :param orientation: The crystal orientation.
        :param n_int: normal to the plane of intersection (laboratory local frame).
        :param view_up: vector to place upwards on the plot.
        :param int trace_size: size of the trace.
        :param verbose: activate verbose mode.
        :return: a numpy array with the coordinates of the two points defining the trace.
        """
        gt = orientation.orientation_matrix().transpose()
        n_rot = gt.dot(self.normal())
        trace_xyz = np.cross(n_rot, n_int)
        trace_xyz /= np.linalg.norm(trace_xyz)
        # now we have the trace vector expressed in the XYZ coordinate system
        # we need to change the coordinate system to the intersection plane
        # (then only the first two component will be non zero)
        P = np.zeros((3, 3), dtype=np.float)
        Zp = n_int
        Yp = view_up / np.linalg.norm(view_up)
        Xp = np.cross(Yp, Zp)
        for k in range(3):
            P[k, 0] = Xp[k]
            P[k, 1] = Yp[k]
            P[k, 2] = Zp[k]
        trace = trace_size * P.transpose().dot(trace_xyz)  # X'=P^-1.X
        if verbose:
            print('n_rot = %s' % n_rot)
            print('trace in XYZ', trace_xyz)
            print(P)
            print('trace in (XpYpZp):', trace)
        return trace

    @staticmethod
    def plot_slip_traces(orientation, hkl='111', n_int=np.array([0, 0, 1]), \
                         view_up=np.array([0, 1, 0]), verbose=False, title=True, legend=True, \
                         trans=False, str_plane=None):
        """
        A method to plot the slip planes intersection with a particular plane
        (known as slip traces if the plane correspond to the surface).
        A few parameters can be used to control the plot looking.
        Thank to Jia Li for starting this code.

        :param orientation: The crystal orientation.
        :param hkl: the slip plane family (eg. 111 or 110)
        :param n_int: normal to the plane of intersection.
        :param view_up: vector to place upwards on the plot.
        :param verbose: activate verbose mode.
        :param title: display a title above the plot.
        :param legend: display the legend.
        :param trans: use a transparent background for the figure (useful to overlay the figure on top of another image).
        :param str_plane: particular string to use to represent the plane in the image name.
        """
        plt.figure()
        hkl_planes = HklPlane.get_family(hkl)
        colors = 'rgykcmbw'
        for i, hkl_plane in enumerate(hkl_planes):
            trace = hkl_plane.slip_trace(orientation, n_int=n_int, view_up=view_up, trace_size=1, verbose=verbose)
            x = [-trace[0] / 2, trace[0] / 2]
            y = [-trace[1] / 2, trace[1] / 2]
            plt.plot(x, y, colors[i % len(hkl_planes)], label='%d%d%d' % hkl_plane.miller_indices(), linewidth=2)
        plt.axis('equal')
        t = np.linspace(0., 2 * np.pi, 100)
        plt.plot(0.5 * np.cos(t), 0.5 * np.sin(t), 'k')
        plt.axis([-0.51, 0.51, -0.51, 0.51])
        plt.axis('off')
        if not str_plane: str_plane = '(%.1f, %.1f, %.1f)' % (n_int[0], n_int[1], n_int[2])
        if title:
            plt.title('{%s} family traces on plane %s' % (hkl, str_plane))
        if legend: plt.legend(bbox_to_anchor=(0.9, 1), loc=2, borderaxespad=0.)
        plt.savefig('slip_traces_%s_%s.png' % (hkl, str_plane), transparent=trans, format='png')

    @staticmethod
    def plot_XY_slip_traces(orientation, hkl='111', title=True,
                            legend=True, trans=False, verbose=False):
        """Helper method to plot the slip traces on the XY plane."""
        HklPlane.plot_slip_traces(orientation, hkl=hkl, n_int=np.array([0, 0, 1]),
                                  view_up=np.array([0, 1, 0]), title=title, legend=legend,
                                  trans=trans, verbose=verbose, str_plane='XY')

    @staticmethod
    def plot_YZ_slip_traces(orientation, hkl='111', title=True,
                            legend=True, trans=False, verbose=False):
        """Helper method to plot the slip traces on the YZ plane."""
        HklPlane.plot_slip_traces(orientation, hkl=hkl, n_int=np.array([1, 0, 0]),
                                  view_up=np.array([0, 0, 1]), title=title, legend=legend,
                                  trans=trans, verbose=verbose, str_plane='YZ')

    @staticmethod
    def plot_XZ_slip_traces(orientation, hkl='111', title=True,
                            legend=True, trans=False, verbose=False):
        """Helper method to plot the slip traces on the XZ plane."""
        HklPlane.plot_slip_traces(orientation, hkl=hkl, n_int=np.array([0, -1, 0]),
                                  view_up=np.array([0, 0, 1]), title=title, legend=legend,
                                  trans=trans, verbose=verbose, str_plane='XZ')

    @staticmethod
    def indices_from_two_directions(uvw1, uvw2):
        """
        Two crystallographic directions :math:`uvw_1` and :math:`uvw_2`define a unique set of hkl planes. 
        This does not depends on the crystal symmetry.
        
        .. math::

           h = v_1 . w_2 - w_1 . v_2 \\\\
           k = w_1 . u_2 - u_1 . w_2 \\\\
           l = u_1 . v_2 - v_1 . u_2

        :param uvw1: The first instance of the `HklDirection` class.
        :param uvw2: The second instance of the `HklDirection` class.
        :return h, k, l: the miller indices of the `HklPlane` defined by the two directions.
        """
        (u1, v1, w1) = uvw1.miller_indices()
        (u2, v2, w2) = uvw2.miller_indices()
        h = v1 * w2 - w1 * v2
        k = w1 * u2 - u1 * w2
        l = u1 * v2 - v1 * u2
        return h, k, l



import numpy as np
from matplotlib import pyplot as plt, cm
from matplotlib.colors import colorConverter
import matplotlib as mpl
from pymicro.view.vtk_utils import *


def hist(data, nb_bins=256, data_range=(0, 255), show=True, save=False, prefix='data', density=False):
    """Histogram of a data array.

    Compute and plot the gray level histogram of the provided data array.

    *Parameters*

    **data**: the data array to analyse.

    **nb_bins**: the number of bins in the histogram.

    **data_range**: the data range to use in the histogram, (0,255) by default.

    **show**: boolean to display the figure using pyplot (defalut True).

    **save**: boolean to save the figure using pyplot (defalut False).

    **prefix**: a string to use in the file name when saving the histogram
    as an image (defaut 'data').

    **density**: a boolean to control wether the histogram is plotted
    using the probability density function, see numpy histogram function
    for more details (default False).

    .. figure:: _static/HDPE_0001-2_512x512x512_uint8_hist.png
        :width: 600 px
        :height: 400 px
        :alt: HDPE_0001-2_512x512x512_uint8_hist.png
        :align: center

        Gray level histogram computed on a 512x512x512 8 bits image.
    """
    print('computing gray level histogram')
    hist, bin_edges = np.histogram(data, bins=nb_bins, range=data_range, density=density)
    bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])
    plt.figure(figsize=(6, 4))
    if density:
        plt.bar(bin_centers, 100 * hist, width=1, fill=True, color='g', edgecolor='g')
        plt.ylabel('Probability (%)')
    else:
        plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        plt.bar(bin_centers, hist, width=1, fill=True, color='g', edgecolor='g')
        plt.ylabel('Counts')
    plt.xlim(data_range)
    plt.xlabel('Gray level value')
    if save: plt.savefig(prefix + '_hist.png', format='png')
    if show: plt.show()


def flat(img, ref, dark):
    """Apply flat field correction to an image.

    :param np.array img: A 2D array representing the image to correct.
    :param np.array ref: The reference image (without the sample), same shape as the image to correct.
    :param np.array dark: A 2D numpy array representing the dark image (thermal noise of the camera).
    :returns np.array float: the flat field corrected image (between 0 and 1) as a float32 numpy array.
    """
    flat = (img - dark).astype(np.float32) / (ref - dark).astype(np.float32)
    return flat


def min_max_cumsum(data, cut=0.001, verbose=False):
    """Compute the indices corresponding to the edge of a distribution.
    
    The min and max indices are calculated according to a cutoff value (0.001 by default) meaning 99.99% 
    of the distribution lies within this range.
    
    """
    total = np.sum(data)
    p = np.cumsum(data)
    nb_bins = data.shape[0]

    mini = 0.0
    maxi = 0.0
    for i in range(nb_bins):
        if p[i] > total * cut:
            mini = i
            if verbose:
                print('min = %f (i=%d)' % (mini, i))
            break
    for i in range(nb_bins):
        if total - p[nb_bins - 1 - i] > total * cut:
            maxi = nb_bins - 1 - i
            if verbose:
                print('max = %f (i=%d)' % (maxi, i))
            break
    return mini, maxi


def auto_min_max(data, cut=0.0002, nb_bins=256, verbose=False):
    """Compute the min and max values in a numpy data array.

    The min and max values are calculated based on the histogram of the
    array which is limited at both ends by the cut parameter (0.0002 by
    default). This means 99.98% of the values are within the [min, max]
    range.

    :param data: the data array to analyse.
    :param float cut: the cut off value to use on the histogram (0.0002 by default).
    :param int nb_bins: number of bins to use in the histogram (256 by default).
    :param bool verbose: activate verbose mode (False by default).
    :returns tuple (val_mini, val_maxi): a tuple containing the min and max values.
    """
    n, bins = np.histogram(data, bins=nb_bins, density=False)
    mini, maxi = min_max_cumsum(n, cut=cut, verbose=verbose)
    val_mini, val_maxi = bins[mini], bins[maxi]
    return val_mini, val_maxi


def region_growing(data, seed, min_thr, max_thr, structure=None):
    """Simple region growing segmentation algorithm.

    This function segment a part of an image by selecting the connected region within a given gray value range.
    Underneath, the method uses the label function of the ndimage package.

    :param data: the numpy data array to segment.
    :param tuple seed: the seed point.
    :param float min_thr: the lower bound of the gray value range.
    :param float max_thr: the upper bound of the gray value range.
    :param structure: structuring element, to use (None by default).
    :returns region: a bool data array with True values for the segmented region.
    """
    from scipy import ndimage
    assert len(seed) == data.ndim
    data[tuple(seed)] = min_thr
    thr_img = (data < max_thr) & (data >= min_thr)
    labels, nb_labels = ndimage.label(thr_img, structure=structure)
    the_label = labels[tuple(seed)]
    region = labels == the_label
    return region


def recad(data, mini, maxi):
    """Cast a numpy array into 8 bit data type.

    This function change the data type of a numpy array into 8 bit. The
    data values are interpolated from [min, max] to [0, 255]. Both min
    and max values may be chosen manually or computed by the function
    `auto_min_max`.

    :param data: the data array to cast to uint8.
    :param float mini: value to use as the minimum (will be 0 in the casted array).
    :param float maxi: value to use as the maximum (will be 255 in the casted array).
    :returns data_uint8: the data array casted to uint8.
    """
    return (255 * (data.clip(mini, maxi) - mini) / (maxi - mini)).astype(np.uint8)


def alpha_cmap(color='red', opacity=1.0):
    """Creating a particular colormap with transparency.

    Only values equal to 255 will have a non zero alpha channel.
    This is typically used to overlay a binary result on initial data.

    :param color: the color to use for non transparent values (ie. 255).
    :param float opacity: opacity value to use for visible pixels.
    :returns mycmap: a fully transparent colormap except for 255 values.
    """
    color1 = colorConverter.to_rgba('white')
    color2 = colorConverter.to_rgba(color)
    mycmap = mpl.colors.LinearSegmentedColormap.from_list('my_cmap', [color1, color2], 256)
    mycmap._init()
    alphas = np.zeros(mycmap.N + 3)
    alphas[255:] = opacity  # show only values at 255
    mycmap._lut[:, -1] = alphas
    return mycmap


def stitch(image_stack, nh=2, nv=1, pattern='E', hmove=None, vmove=None, adjust_bc=False, adjust_bc_nbins=256,
           verbose=False, show=True, save=False, save_name='stitch', save_ds=1, check=None):
    """Stich a series of images together.
    
    :param list image_stack: a list of the images to stitch.
    :param int nh: number of images to stitch horizontally.
    :param int nv: number of images to stitch vertically.
    :param str pattern: stitching pattern ('E' or 'S').
    :param tuple hmove: horizontal move between 2 images (first image width by default).
    :param tuple vmove: vertical move between 2 images (first image height by default).
    :param bool adjust_bc: adjust gray levels by comparing the histograms in the overlapping regions (False by default).
    :param int adjust_bc_nbins: numbers of bins to use tin the histograms when adjusting the gray levels (256 by default).
    :param bool verbose: activate verbose mode (False by default).
    :param bool show: show the stitched image (True by default).
    :param bool save: flag to save the image as png to the disk (False by default).
    :param str save_name: name to use to save the stitched image (stitch by default).
    :param int save_ds: downsampling factor using when saving the stitched image (1 by default).
    :param int check: trigger plotting for the given image if set.
    :returns full_im: the stitched image in [x, y] form.
    """
    assert pattern in ['E', 'S']
    # is motion not set, assume the full size of the first image
    if hmove is None:
        hmove = (image_stack[0].shape[0], 0)
    if vmove is None:
        vmove = (0, image_stack[0].shape[1])
    H = np.array(hmove).astype(int)
    V = np.array(vmove).astype(int)
    image_size = np.array(image_stack[0].shape)
    if adjust_bc:
        # make sure the images are overlapping
        assert image_size[0] > H[0]
        assert image_size[1] > V[1]
        bch = image_size[0] - H[0]  # horizontal overlap
        bcv = image_size[1] - V[1]  # vertical overlap
        if verbose:
            print('overlapping is {0} pixels horizontally and {1} pixels vertically'.format(bch, bcv))
    data_type = image_stack[0].dtype
    full_im = np.zeros((image_size[0] + (nh - 1) * H[0] + nv * V[0],
                        image_size[1] + (nv - 1) * V[1] + nh * H[1]),
                       dtype=data_type)
    if verbose:
        print('image_size is {0}'.format(image_size))
        print('image data type is {0}'.format(data_type))
        print('full image_size will be {0}'.format(full_im.shape))
    for i in range(len(image_stack)):
        # compute indices
        x = i % nh
        y = i // nh
        if pattern == 'S' and (y % 2) == 1:
            x = nh - 1 - x
        im = image_stack[i]
        xp, yp = x * H + y * V + image_size // 2
        if adjust_bc:
            if i == 0:
                # this is the reference
                pass
            else:
                # isolate the overlapping regions
                if i % nh != 0:
                    x1s = xp - image_size[0] // 2
                    x1e = xp - image_size[0] // 2 + bch
                    y1s = yp - image_size[1] // 2
                    y1e = yp - image_size[1] // 2 + image_size[1]
                    region1 = full_im[x1s:x1e, y1s:y1e]
                    region2 = im[:bch, :]
                else:
                    x1s = xp - image_size[0] // 2
                    x1e = xp - image_size[0] // 2 + image_size[0]
                    y1s = yp - image_size[1] // 2
                    y1e = yp - image_size[1] // 2 + bcv
                    region1 = full_im[x1s:x1e, y1s:y1e]
                    region2 = im[:, :bcv]
                if i == check:
                    fig = plt.figure()
                    ax1 = fig.add_subplot(121 if i % nh != 0 else 211)
                    im1 = ax1.imshow(region1.T, cmap=cm.gray, vmin=0, vmax=65535)
                    fig.colorbar(im1, ax=ax1)
                    ax1.set_title('region1')
                    ax2 = fig.add_subplot(122 if i % nh != 0 else 212)
                    im2 = ax2.imshow(region2.T, cmap=cm.gray, vmin=0, vmax=65535)
                    fig.colorbar(im2, ax=ax2)
                    ax2.set_title('region2')
                    plt.show()
                region1 = region1.flatten()
                region2 = region2.flatten()
                # work out the histograms
                data_type_min = np.iinfo(data_type).min
                data_type_max = np.iinfo(data_type).max
                hist1, bin_edges = np.histogram(region1, bins=adjust_bc_nbins, density=False,
                                                range=(data_type_min, data_type_max))
                hist2, bin_edges = np.histogram(region2, bins=adjust_bc_nbins, density=False,
                                                range=(data_type_min, data_type_max))
                if i == check:
                    hist(region1, data_range=(data_type_min, data_type_max), nb_bins=adjust_bc_nbins)
                    hist(region2, data_range=(data_type_min, data_type_max), nb_bins=adjust_bc_nbins)
                # compute the difference not taking into account the edges of the histograms
                max1 = np.argmax(hist1[1:-2])
                max2 = np.argmax(hist2[1:-2])
                diff = (data_type_max - data_type_min) / adjust_bc_nbins * (max1 - max2)
                if verbose:
                    print('matching max of histograms in both images: {0:d} and {1:d}, adding {2:d}'.format(max1, max2, diff))
                # keep a reference of saturated value (we do not want to change those)
                im_is_min = im == data_type_min
                im_is_max = im == data_type_max
                # shift all the gray levels, making sure not to overflow the image type range
                im = im.astype(float) + diff
                im[im < data_type_min] = data_type_min
                im[im > data_type_max] = data_type_max
                im = im.astype(data_type)
                # reset saturated values
                im[im_is_min] = data_type_min
                im[im_is_max] = data_type_max
        full_im[xp - image_size[0] // 2:xp - image_size[0] // 2 + image_size[0],
            yp - image_size[1] // 2:yp - image_size[1] // 2 + image_size[1]] = im
    if save:
        plt.imsave('{0:s}.png'.format(save_name), full_im[::save_ds, ::save_ds].T,
                   cmap=cm.gray)
    if show:
        plt.imshow(full_im[::save_ds, ::save_ds].T, interpolation='nearest', cmap=cm.gray)
        plt.axis('off')
        plt.show()
    return full_im


def compute_affine_transform(fixed, moving):
    """Compute the affine transform by point set registration.

    The affine transform is the composition of a translation and a linear map.
    The two ordered lists of points must be of the same length larger or equal to 3.
    The order of the points in the two list must match.
    
    The 2D affine transform :math:`\mathbf{A}` has 6 parameters (2 for the translation and 4 for the linear transform).
    The best estimate of :math:`\mathbf{A}` can be computed using at least 3 pairs of matching points. Adding more 
    pair of points will improve the quality of the estimate. The matching pairs are usually obtained by selecting 
    unique features in both images and measuring their coordinates.

    :param list fixed: a list of the reference points.
    :param list moving: a list of the moving points to register on the fixed point.
    :returns translation, linear_map: the computed translation and linear map affine transform. 

    Thanks to Will Lenthe for helping with this code.
    """
    assert len(fixed) == len(moving)
    assert len(fixed) >= 3
    fixed_centroid = np.average(fixed, 0)
    moving_centroid = np.average(moving, 0)
    # offset every point by the center of mass of all the points in the set
    fixed_from_centroid = fixed - fixed_centroid
    moving_from_centroid = moving - moving_centroid
    covariance = moving_from_centroid.T.dot(fixed_from_centroid)
    variance = moving_from_centroid.T.dot(moving_from_centroid)
    # compute the full affine transform: translation + linear map
    linear_map = np.linalg.inv(variance).dot(covariance).T
    translation = fixed_centroid - linear_map.dot(moving_centroid)
    return translation, linear_map



import os, numpy as np
from matplotlib import pyplot as plt
from skimage.transform import radon
from math import *
from config import PYMICRO_XRAY_DATA_DIR

densities = {'Li': 0.533,  # Z = 3
             'Be': 1.8450,  # Z = 4
             'C': 2.26,  # Z = 6, strongly depends on crystal structure
             'Mg': 1.738,  # Z = 12
             'Al': 2.6941,  # Z = 13
             'Si': 2.33,  # Z = 14
             'Ti': 4.530,  # Z = 22
             'V': 6.100,  # Z = 23
             'Cr': 7.180,  # Z = 24
             'Mn': 7.430,  # Z = 25
             'Fe': 7.874,  # Z = 26
             'Co': 8.900,  # Z = 27
             'Ni': 8.902,  # Z = 28
             'Cu': 8.960,  # Z = 29
             'Zn': 7.112,  # Z = 30
             'Ga': 7.877,  # Z = 31
             'Ge': 5.370,  # Z = 32
             'Nb': 8.550,  # Z = 41
             'Pb': 11.330,  # Z = 82
             'WC': 15.63, # Z(W) = 74
             }


def f_atom(q, Z):
    """Empirical function for the atomic form factor f.

    This function implements the empirical function from the paper of Muhammad and Lee
    to provide value of f for elements in the range Z <= 30.
    doi:10.1371/journal.pone.0069608.t001

    :param float q: value or series for the momentum transfer, unit is angstrom^-1
    :param int Z: atomic number, must be lower than 30.
    :return: the atomic form factor value or series corresponding to q.
    """
    if Z > 30:
        raise ValueError('only atoms with Z<=30 are supported, consider using tabulated data.')
    a0 = 0.52978  # angstrom, Bohr radius
    params = np.genfromtxt(os.path.join(PYMICRO_XRAY_DATA_DIR, 'f_atom_params.txt'), names=True)
    Z, r, a1, b1, a2, b2, a3, b3, _, _, _ = params[int(Z - 1)]
    print(Z, r, a1, b1, a2, b2, a3, b3)
    f = (a1 * Z) ** r / ((a1 * Z) ** r + b1 * (2 * pi * a0 * q) ** r) ** r + \
        (a2 * Z) ** r / ((2 * a2 * Z) ** r + b2 * (2 * pi * a0 * q) ** 2) ** 2 + \
        (a3 * Z) ** r / ((2 * a3 * Z) ** 2 + b3 * (2 * pi * a0 * q) ** 2) ** 2
    return f


def atom_scat_factor_function(mat='Al', sintheta_lambda_max=12, display=True):
    """Compute and display the fit function of the atomic scattering factor.

    :param string mat: A string representing the material (e.g. 'Al')
    :param float sintheta_lambda_max: maximal value of sin theta / lambda
    :param bool display: display an image of the plot
    """
    param = np.genfromtxt(os.path.join(PYMICRO_XRAY_DATA_DIR, mat + '_fit_fatom'))
    print('Fit coefficient for', param[:,1])
    fit = []
    sintheta_lambda = np.linspace(0.0, sintheta_lambda_max, 100)

    for i in sintheta_lambda:
        fit_fatom = param[0, 1] * exp(-param[1, 1] * i ** 2) + param[2, 1] * exp(-param[3, 1] * i ** 2) + param[4, 1] * exp(-param[5, 1] * i ** 2) + param[6, 1] * exp(-param[7, 1] * i ** 2) + param[8, 1]
        fit.append(fit_fatom)
    print(fit)

    plt.figure()
    plt.plot(sintheta_lambda, fit, 'o-', label=mat)
    plt.legend(loc='upper right')
    plt.xlabel(r'$\sin\theta / \lambda (nm^{-1})$')
    plt.ylabel('Atomic scattering factor')
    plt.title('Fit function of : %s' % mat)
    if display:
        plt.show()
    return fit


def lambda_keV_to_nm(lambda_keV):
    """Change the unit of wavelength from keV to nm.

    :param float lambda_keV: the wavelength in keV unit.
    :returns: the wavelength in nm unit.
    """
    return 1.2398 / lambda_keV


def lambda_keV_to_angstrom(lambda_keV):
    """Change the unit of wavelength from keV to angstrom.

    :param float lambda_keV: the wavelength in keV unit.
    :returns: the wavelength in angstrom unit.
    """
    return 12.398 / lambda_keV


def lambda_nm_to_keV(lambda_nm):
    """Change the unit of wavelength from nm to keV.

    :param float lambda_nm: the wavelength in nm unit.
    :returns: the wavelength in keV unit.
    """
    return 1.2398 / lambda_nm


def lambda_angstrom_to_keV(lambda_angstrom):
    """Change the unit of wavelength from angstrom to keV.

    :param float lambda_angstrom: the wavelength in angstrom unit.
    :returns: the wavelength in keV unit.
    """
    return 12.398 / lambda_angstrom


def plot_xray_trans(mat='Al', ts=[1.0], rho=None, energy_lim=[1, 100], legfmt='%.1f', display=True):
    """Plot the transmitted intensity of a X-ray beam through a given material.

    This function compute the transmitted intensity from tabulated data of
    the mass attenuation coefficient \mu_\rho (between 1 and 100 keV) and
    applying Beer's Lambert law:

    .. math::

      I/I_0 = \exp(-\mu_\rho*\rho*t)

    The tabulated data is stored in ascii files in the data folder. It has been retrieved
    from NIST `http://physics.nist.gov/cgi-bin/ffast/ffast.pl`
    The density is also tabulated and can be left blanked unless a specific value is to be used.

    :param string mat: A string representing the material (e.g. 'Al')
    :param list ts: a list of thickness values of the material in mm ([1.0] by default)
    :param float rho: density of the material in g/cm^3 (None by default)
    :param list energy_lim: energy bounds in keV for the plot (1, 100 by default)
    :param string legfmt: string to format the legend plot
    :param bool display: display or save an image of the plot (False by default)
    """
    mu_rho = np.genfromtxt(os.path.join(PYMICRO_XRAY_DATA_DIR, mat + '.txt'), usecols=(0, 1), comments='#')
    print('Data :', mu_rho)
    energy = mu_rho[:, 0]
    print('Energy :', energy)
    # look up density
    if rho is None:
        rho = densities[mat]
    legstr = '%%s %s mm' % legfmt
    plt.figure()
    for t in ts:
        # apply Beer-Lambert
        trans = 100 * np.exp(-mu_rho[:, 1] * rho * t / 10)
        plt.plot(energy, trans, '-', linewidth=3, markersize=10, label=legstr % (mat, t))
    # bound the energy to (1, 200)
    if energy_lim[0] < 1:
        energy_lim[0] = 1
    if energy_lim[1] > 200:
        energy_lim[1] = 200
    plt.xlim(energy_lim)
    #plt.ylim(0, 10)
    plt.grid()
    plt.legend(loc='upper left')
    plt.xlabel('Photon Energy (keV)')
    plt.ylabel('Transmission I/I0 (%)')
    plt.title('Transmitted intensity of : %s' % mat)
    if display:
        plt.show()
    else:
        plt.savefig('xray_trans_' + mat + '.png')

def radiograph(data, omega):
    """Compute a single radiograph of a 3D object using the radon transform.

    :param np.array data: an array representing the 3D object in (XYZ) form.
    :param omega: the rotation angle value in degrees.
    :returns projection: a 2D array in (Y, Z) form.
    """
    projection = radiographs(data, [omega])
    return projection[:, :, 0]

def radiographs(data, omegas):
    """Compute the radiographs of a 3D object using the radon transform.

    The object is represented by a 3D numpy array in (XYZ) form and a series of projection at each omega angle
    are computed assuming the rotation is along Z in the middle of the data set. Internally this function uses
    the radon transform from the skimage package.

    :param np.array data: an array representing the 3D object in (XYZ) form.
    :param omegas: an array of the rotation values in degrees.
    :returns projections: a 3D array in (Y, Z, omega) form.
    """
    assert data.ndim == 3
    if type(omegas) is list:
        omegas = np.array(omegas)
    width = int(np.ceil(max(data.shape[0], data.shape[1]) * 2 ** 0.5))
    projections = np.zeros((width, np.shape(data)[2], len(omegas)), dtype=np.float)
    for z in range(np.shape(data)[2]):
        a = radon(data[:, :, z], -omegas, circle=False)  # - 90  # the 90 seems to come from the radon function itself
        projections[:, z, :] = a[:, :]
    return projections



